"""Neural network modules using fully connected hidden layers."""
from typing import Tuple

import torch
import torch.nn as nn
from ray.rllib.utils import override

from .linear import MaskedLinear
from .utils import get_activation


class FullyConnected(nn.Sequential):
    """Applies several fully connected modules to inputs."""

    def __init__(
        self,
        in_features: int,
        units: Tuple[int, ...] = (),
        activation: str = None,
        layer_norm: bool = False,
    ):
        super().__init__()
        self.in_features = in_features
        activ = get_activation(activation)
        units = (self.in_features,) + tuple(units)
        modules = []
        for in_dim, out_dim in zip(units[:-1], units[1:]):
            modules.append(nn.Linear(in_dim, out_dim))
            if layer_norm:
                modules.append(nn.LayerNorm(out_dim))
            if activ:
                modules.append(activ())
        self.out_features = units[-1]
        self.sequential = nn.Sequential(*modules)

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        # pylint:disable=arguments-differ
        return self.sequential(inputs)


class StateActionEncoder(nn.Module):
    """Concatenates action after the first layer."""

    __constants__ = {"in_features", "out_features"}

    def __init__(
        self,
        obs_dim: int,
        action_dim: int,
        delay_action: bool = True,
        units: Tuple[int, ...] = (),
        **fc_kwargs
    ):
        super().__init__()
        self.in_features = obs_dim + action_dim
        if units:
            if delay_action is True:
                self.obs_module = FullyConnected(obs_dim, units=units[:1], **fc_kwargs)
                input_dim = units[0] + action_dim
                units = units[1:]
            else:
                self.obs_module = nn.Identity()
                input_dim = obs_dim + action_dim
            self.sequential_module = FullyConnected(input_dim, units=units, **fc_kwargs)
            self.out_features = self.sequential_module.out_features
        else:
            self.obs_module = nn.Identity()
            self.sequential_module = nn.Identity()
            self.out_features = obs_dim + action_dim

    @override(nn.Module)
    def forward(self, obs: torch.Tensor, actions: torch.Tensor) -> torch.Tensor:
        # pylint:disable=arguments-differ
        output = self.obs_module(obs)
        output = torch.cat([output, actions], dim=-1)
        output = self.sequential_module(output)
        return output


class MADE(nn.Module):
    """MADE: Masked Autoencoder for Distribution Estimation

    Implements a Masked Autoregressive MLP, where carefully constructed
    binary masks over weights ensure the autoregressive property.

    Based on: https://github.com/karpathy/pytorch-made

    Args:
        in_features: number of inputs
        hidden sizes: number of units in hidden layers
        out_features: number of outputs, which usually collectively parameterize
            some kind of 1D distribution
        natural_ordering: force natural ordering of dimensions,
            don't use random permutations

    Note: if out_features is e.g. 2x larger than `in_features` (perhaps the mean
        and std), then the first `in_features` outputs will be all the means and
        the remaining will be stds. I.e. output dimensions depend on the same
        input dimensions in "chunks" and should be carefully decoded downstream.
    """

    # pylint:disable=too-many-instance-attributes

    def __init__(
        self,
        in_features: int,
        units: Tuple[int, ...],
        out_features: int,
        natural_ordering: bool = False,
    ):
        # pylint:disable=too-many-arguments
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        assert (
            self.out_features % self.in_features == 0
        ), "out_features must be integer multiple of in_features"

        # define a simple MLP neural net
        sizes = [in_features] + list(units) + [out_features]
        # define input units ids
        ids = [torch.arange(sizes[0]) if natural_ordering else torch.randperm(sizes[0])]
        # define hidden units ids
        for idx, size in enumerate(sizes[1:-1]):
            ids.append(torch.randint(ids[idx].min().item(), out_features - 1, (size,)))
        # copy output units ids from input units ids
        ids.append(torch.cat([ids[0]] * (out_features // in_features), dim=-1))
        # define masks for each layer
        masks = [m.unsqueeze(-1) >= n.unsqueeze(0) for m, n in zip(ids[1:-1], ids[:-2])]
        # last layer has a different connection pattern
        masks.append(ids[-1].unsqueeze(-1) > ids[-2].unsqueeze(0))

        linears = [MaskedLinear(hin, hout) for hin, hout in zip(sizes[:-1], sizes[1:])]
        for linear, mask in zip(linears, masks):
            linear.set_mask(mask)
        layers = [m for layer in linears[:-1] for m in (layer, nn.LeakyReLU(0.2))]
        layers += linears[-1:]
        self.net = nn.Sequential(*layers)

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.net(inputs)



"""Customized modules."""
from .action_output import ActionOutput
from .activation import Swish
from .dist_params import (
    CategoricalParams,
    NormalParams,
    PolicyNormalParams,
    StdNormalParams,
)
from .fully_connected import MADE, FullyConnected, StateActionEncoder
from .gaussian_noise import GaussianNoise
from .lambd import Lambda
from .leaf_parameter import LeafParameter
from .linear import MaskedLinear, NormalizedLinear
from .tanh_squash import TanhSquash
from .tril_matrix import TrilMatrix

__all__ = [
    "ActionOutput",
    "Swish",
    "CategoricalParams",
    "LeafParameter",
    "FullyConnected",
    "GaussianNoise",
    "Lambda",
    "NormalParams",
    "PolicyNormalParams",
    "NormalizedLinear",
    "MADE",
    "MaskedLinear",
    "StateActionEncoder",
    "StdNormalParams",
    "TanhSquash",
    "TrilMatrix",
]



"""Base classes for distribution modules."""
from typing import Dict, List

import numpy as np
import torch
import torch.nn as nn
from ray.rllib.utils import override

from . import flows


class ConditionalDistribution(nn.Module):
    """Implements torch.distribution.Distribution interface as a nn.Module.

    If passed a Distribution, wraps the unconditional distribution to be used with the
    ConditionalDistribution interface.
    """

    # pylint:disable=abstract-method,unused-argument,not-callable

    def __init__(self, *, distribution: nn.Module = None):
        super().__init__()
        self.distribution = distribution

    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        if self.distribution is not None:
            return self.distribution.sample(sample_shape)
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()

    @torch.jit.export
    def rsample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        if self.distribution is not None:
            return self.distribution.rsample(sample_shape)
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()

    @torch.jit.export
    def log_prob(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        """
        Returns the log of the probability density/mass function evaluated at `value`.
        """
        if self.distribution is not None:
            return self.distribution.log_prob(value)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def cdf(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        """Returns the cumulative density/mass function evaluated at `value`."""
        if self.distribution is not None:
            return self.distribution.cdf(value)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def icdf(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        """Returns the inverse cumulative density/mass function evaluated at `value`."""
        if self.distribution is not None:
            return self.distribution.icdf(value)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def entropy(self, params: Dict[str, torch.Tensor]):
        """Returns entropy of distribution."""
        if self.distribution is not None:
            return self.distribution.entropy()
        return torch.tensor(np.nan).float()

    @torch.jit.export
    def perplexity(self, params: Dict[str, torch.Tensor]):
        """Returns perplexity of distribution."""
        return self.entropy(params).exp()

    @torch.jit.export
    def reproduce(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        """Produce a reparametrized sample with the same value as `value`."""
        if self.distribution is not None:
            return self.distribution.reproduce(value)
        return (
            torch.tensor(np.nan).float().expand_as(value),
            torch.tensor(np.nan).float().expand_as(value),
        )

    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        """
        Generates a deterministic sample or batch of samples if the distribution
        parameters are batched. Returns a (rsample, log_prob) pair.
        """
        if self.distribution is not None:
            return self.distribution.deterministic()
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()


class Distribution(nn.Module):
    """Unconditional Distribution.

    If passed a ConditionalDistribution, wraps the unconditional distribution to be used
    with the Distribution interface. `nn.Parameter`s passed as distribution parameters
    will be registered as module parameters, making the distribution learnable.
    Otherwise, parameters will be registered as buffers.
    """

    # pylint:disable=abstract-method,not-callable
    params: Dict[str, torch.Tensor]

    def __init__(
        self, *, cond_dist: nn.Module = None, params: Dict[str, torch.Tensor] = None
    ):
        super().__init__()
        self.cond_dist = cond_dist
        self.params = params or {}
        for name, param in self.params.items():
            if isinstance(param, nn.Parameter):
                self.register_parameter(name, param)
            else:
                self.register_buffer(name, param)

    @torch.jit.export
    def sample(self, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        if self.cond_dist is not None:
            return self.cond_dist.sample(self.params, sample_shape)
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()

    @torch.jit.export
    def rsample(self, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        if self.cond_dist is not None:
            return self.cond_dist.rsample(self.params, sample_shape)
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()

    @torch.jit.export
    def log_prob(self, value):
        """
        Returns the log of the probability density/mass function evaluated at `value`.
        """
        if self.cond_dist is not None:
            return self.cond_dist.log_prob(value, self.params)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def cdf(self, value):
        """Returns the cumulative density/mass function evaluated at `value`."""
        if self.cond_dist is not None:
            return self.cond_dist.cdf(value, self.params)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def icdf(self, value):
        """Returns the inverse cumulative density/mass function evaluated at `value`."""
        if self.cond_dist is not None:
            return self.cond_dist.icdf(value, self.params)
        return torch.tensor(np.nan).float().expand_as(value)

    @torch.jit.export
    def entropy(self):
        """Returns entropy of distribution."""
        if self.cond_dist is not None:
            return self.cond_dist.entropy(self.params)
        return torch.tensor(np.nan).float()

    @torch.jit.export
    def perplexity(self):
        """Returns perplexity of distribution."""
        return self.entropy().exp()

    @torch.jit.export
    def reproduce(self, value):
        """Produce a reparametrized sample with the same value as `value`."""
        if self.cond_dist is not None:
            return self.cond_dist.reproduce(value, self.params)
        return (
            torch.tensor(np.nan).float().expand_as(value),
            torch.tensor(np.nan).float().expand_as(value),
        )

    @torch.jit.export
    def deterministic(self):
        """
        Generates a deterministic sample or batch of samples if the distribution
        parameters are batched. Returns a (rsample, log_prob) pair.
        """
        if self.cond_dist is not None:
            return self.cond_dist.deterministic(self.params)
        return torch.tensor(np.nan).float(), torch.tensor(np.nan).float()


class Independent(ConditionalDistribution):
    """Reinterprets some of the batch dims of a distribution as event dims."""

    # pylint:disable=abstract-method

    def __init__(self, base_dist, reinterpreted_batch_ndims):
        super().__init__()
        self.base_dist = base_dist
        self.reinterpreted_batch_ndims = reinterpreted_batch_ndims

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.base_dist(inputs)

    @override(ConditionalDistribution)
    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out, base_log_prob = self.base_dist.sample(params, sample_shape)
        return (
            out,
            flows.utils.sum_rightmost(base_log_prob, self.reinterpreted_batch_ndims),
        )

    @override(ConditionalDistribution)
    @torch.jit.export
    def rsample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out, base_log_prob = self.base_dist.rsample(params, sample_shape)
        if out is not None:
            return (
                out,
                flows.utils.sum_rightmost(
                    base_log_prob, self.reinterpreted_batch_ndims
                ),
            )
        return out, base_log_prob

    @override(ConditionalDistribution)
    @torch.jit.export
    def log_prob(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        base_log_prob = self.base_dist.log_prob(value, params)
        return flows.utils.sum_rightmost(base_log_prob, self.reinterpreted_batch_ndims)

    @override(ConditionalDistribution)
    @torch.jit.export
    def cdf(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        return self.base_dist.cdf(value, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def icdf(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        return self.base_dist.icdf(value, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def entropy(self, params: Dict[str, torch.Tensor]):
        base_entropy = self.base_dist.entropy(params)
        return flows.utils.sum_rightmost(base_entropy, self.reinterpreted_batch_ndims)

    @override(ConditionalDistribution)
    @torch.jit.export
    def reproduce(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        sample_, log_prob_ = self.base_dist.reproduce(value, params)
        return (
            sample_,
            flows.utils.sum_rightmost(log_prob_, self.reinterpreted_batch_ndims),
        )

    @override(ConditionalDistribution)
    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        sample, log_prob = self.base_dist.deterministic(params)
        return (
            sample,
            flows.utils.sum_rightmost(log_prob, self.reinterpreted_batch_ndims),
        )


class TransformedDistribution(ConditionalDistribution):
    """
    Extension of the ConditionalDistribution class, which applies a sequence of
    transformations to a base distribution.
    """

    # pylint:disable=abstract-method

    def __init__(self, base_dist, transform):
        super().__init__()
        self.base_dist = (
            ConditionalDistribution(distribution=base_dist)
            if isinstance(base_dist, Distribution)
            else base_dist
        )
        self.transform = (
            flows.ConditionalTransform(transform=transform)
            if isinstance(transform, flows.Transform)
            else transform
        )

    @override(ConditionalDistribution)
    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        base_sample, base_log_prob = self.base_dist.sample(params, sample_shape)
        transformed, log_abs_det_jacobian = self.transform(base_sample, params)
        return transformed.detach(), base_log_prob - log_abs_det_jacobian

    @override(ConditionalDistribution)
    @torch.jit.export
    def rsample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        base_rsample, base_log_prob = self.base_dist.rsample(params, sample_shape)
        transformed, log_abs_det_jacobian = self.transform(base_rsample, params)
        return transformed, base_log_prob - log_abs_det_jacobian

    @override(ConditionalDistribution)
    @torch.jit.export
    def log_prob(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        latent, log_abs_det_jacobian = self.transform(value, params, reverse=True)
        base_log_prob = self.base_dist.log_prob(latent, params)
        return base_log_prob + log_abs_det_jacobian

    @override(ConditionalDistribution)
    @torch.jit.export
    def reproduce(self, value: torch.Tensor, params: Dict[str, torch.Tensor]):
        latent, _ = self.transform(value, params, reverse=True)
        latent_, base_log_prob_ = self.base_dist.reproduce(latent, params)
        value_, log_abs_det_jacobian_ = self.transform(latent_, params)
        return value_, base_log_prob_ - log_abs_det_jacobian_

    @override(ConditionalDistribution)
    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        base_sample, base_log_prob = self.base_dist.deterministic(params)
        transformed, log_abs_det_jacobian = self.transform(base_sample, params)
        return transformed, base_log_prob - log_abs_det_jacobian



"""Distributions as PyTorch modules compatible with TorchScript."""
from . import flows, types
from .abstract import (
    ConditionalDistribution,
    Distribution,
    Independent,
    TransformedDistribution,
)
from .categorical import Categorical
from .normal import Normal
from .uniform import Uniform

__all__ = [
    "flows",
    "types",
    "ConditionalDistribution",
    "Distribution",
    "Independent",
    "TransformedDistribution",
    "Categorical",
    "Normal",
    "Uniform",
]



"""Base for all PyTorch policies."""
import textwrap
from typing import Dict, List, Optional, Set, Tuple, Union

import torch
import torch.nn as nn
from gym.spaces import Space
from ray.rllib import Policy, SampleBatch
from ray.rllib.evaluation.episode import MultiAgentEpisode
from ray.rllib.models.action_dist import ActionDistribution
from ray.rllib.models.modelv2 import flatten, restore_original_dimensions
from ray.rllib.utils import override
from ray.rllib.utils.torch_ops import convert_to_non_torch_type, convert_to_torch_tensor
from ray.rllib.utils.tracking_dict import UsageTrackingDict
from ray.rllib.utils.typing import ModelGradients, TensorType
from ray.tune.logger import pretty_print
from torch import Tensor

from raylab.options import RaylabOptions, configure, option
from raylab.torch.utils import convert_to_tensor
from raylab.utils.types import StatDict, TensorDict

from .compat import WrapRawModule
from .modules import get_module
from .optimizer_collection import OptimizerCollection


@configure
@option("env", default=None)
@option("env_config/", allow_unknown_subkeys=True)
@option("explore", default=True)
@option(
    "exploration_config/", allow_unknown_subkeys=True, override_all_if_type_changes=True
)
@option("framework", default="torch")
@option("gamma", default=0.99)
@option("num_workers", default=0)
@option("seed", default=None)
@option("worker_index", default=0)
@option(
    "module/",
    help="Type and config of the PyTorch NN module.",
    allow_unknown_subkeys=True,
    override_all_if_type_changes=True,
)
@option(
    "optimizer/",
    help="Config dict for PyTorch optimizers.",
    allow_unknown_subkeys=True,
)
@option("compile", False, help="Whether to optimize the policy's backend")
class TorchPolicy(Policy):
    """A Policy that uses PyTorch as a backend.

    Attributes:
        observation_space: Space of possible observation inputs
        action_space: Space of possible action outputs
        config: Policy configuration
        dist_class: Action distribution class for computing actions. Must be set
            by subclasses before calling `__init__`.
        device: Device in which the parameter tensors reside. All input samples
            will be converted to tensors and moved to this device
        module: The policy's neural network module. Should be compilable to
            TorchScript
        optimizers: The optimizers bound to the neural network (or submodules)
        options: Configuration object for this class
    """

    observation_space: Space
    action_space: Space
    config: dict
    global_config: dict
    device: torch.device
    model: WrapRawModule
    module: nn.Module
    optimizers: OptimizerCollection
    options: RaylabOptions = RaylabOptions()

    def __init__(self, observation_space: Space, action_space: Space, config: dict):
        # Allow subclasses to set `dist_class` before calling init
        action_dist = getattr(self, "dist_class", None)
        super().__init__(observation_space, action_space, self._build_config(config))

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.module = self._make_module(observation_space, action_space, self.config)
        self.module.to(self.device)
        self.model = WrapRawModule(
            self.observation_space,
            self.action_space,
            self.module,
            num_outputs=1,
            model_config=self.config["module"],
        )

        self.optimizers = self._make_optimizers()

        # === Policy attributes ===
        self.dist_class = action_dist
        self.dist_class.check_model_compat(self.module)
        self.framework = "torch"  # Needed to create exploration
        self.exploration = self._create_exploration()

    # ==========================================================================
    # PublicAPI
    # ==========================================================================

    @property
    def pull_from_global(self) -> Set[str]:
        """Keys to pull from global configuration.

        Configurations passed down from caller (usually by the trainer) that are
        not under the `policy` config.
        """
        return {
            "env",
            "env_config",
            "explore",
            # "exploration_config",
            "gamma",
            "num_workers",
            "seed",
            "worker_index",
        }

    def compile(self):
        """Optimize modules with TorchScript.

        Warnings:
            This action cannot be undone.
        """
        self.module = torch.jit.script(self.module)

    @torch.no_grad()
    @override(Policy)
    def compute_actions(
        self,
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Union[List[TensorType], TensorType] = None,
        prev_reward_batch: Union[List[TensorType], TensorType] = None,
        info_batch: Optional[Dict[str, list]] = None,
        episodes: Optional[List[MultiAgentEpisode]] = None,
        explore: Optional[bool] = None,
        timestep: Optional[int] = None,
        **kwargs,
    ) -> Tuple[TensorType, List[TensorType], Dict[str, TensorType]]:
        # pylint:disable=too-many-arguments,too-many-locals
        explore = explore if explore is not None else self.config["explore"]
        timestep = timestep if timestep is not None else self.global_timestep

        input_dict = self.lazy_tensor_dict(
            {SampleBatch.CUR_OBS: obs_batch, "is_training": False}
        )
        if prev_action_batch:
            input_dict[SampleBatch.PREV_ACTIONS] = prev_action_batch
        if prev_reward_batch:
            input_dict[SampleBatch.PREV_REWARDS] = prev_reward_batch
        state_batches = convert_to_torch_tensor(state_batches or [], device=self.device)

        # Call the exploration before_compute_actions hook.
        self.exploration.before_compute_actions(timestep=timestep)

        dist_inputs, state_out = self._compute_module_output(
            self._unpack_observations(input_dict),
            state_batches,
            self.convert_to_tensor([1]),
        )

        # pylint:disable=not-callable
        action_dist = self.dist_class(dist_inputs, self.model)
        # pylint:enable=not-callable
        actions, logp = self.exploration.get_exploration_action(
            action_distribution=action_dist, timestep=timestep, explore=explore
        )
        input_dict[SampleBatch.ACTIONS] = actions

        # Add default and custom fetches.
        extra_fetches = self._extra_action_out(
            input_dict, state_batches, self.module, action_dist
        )

        if logp is not None:
            extra_fetches[SampleBatch.ACTION_PROB] = logp.exp()
            extra_fetches[SampleBatch.ACTION_LOGP] = logp

        return convert_to_non_torch_type((actions, state_out, extra_fetches))

    @torch.no_grad()
    @override(Policy)
    def compute_log_likelihoods(
        self,
        actions: Union[List[TensorType], TensorType],
        obs_batch: Union[List[TensorType], TensorType],
        state_batches: Optional[List[TensorType]] = None,
        prev_action_batch: Optional[Union[List[TensorType], TensorType]] = None,
        prev_reward_batch: Optional[Union[List[TensorType], TensorType]] = None,
    ) -> TensorType:
        # pylint:disable=too-many-arguments
        input_dict = self.lazy_tensor_dict(
            {SampleBatch.CUR_OBS: obs_batch, SampleBatch.ACTIONS: actions}
        )
        if prev_action_batch:
            input_dict[SampleBatch.PREV_ACTIONS] = prev_action_batch
        if prev_reward_batch:
            input_dict[SampleBatch.PREV_REWARDS] = prev_reward_batch

        dist_inputs, _ = self.module(
            self._unpack_observations(input_dict),
            state_batches,
            self.convert_to_tensor([1]),
        )
        # pylint:disable=not-callable
        action_dist = self.dist_class(dist_inputs, self.module)
        # pylint:enable=not-callable
        log_likelihoods = action_dist.logp(input_dict[SampleBatch.ACTIONS])
        return log_likelihoods

    @override(Policy)
    def postprocess_trajectory(
        self, sample_batch, other_agent_batches=None, episode=None
    ):
        if not self.config["env_config"].get("time_aware", False):
            hit_limit = sample_batch[SampleBatch.INFOS][-1].get("TimeLimit.truncated")
            env_done = sample_batch[SampleBatch.DONES][-1]
            sample_batch[SampleBatch.DONES][-1] = False if hit_limit else env_done
        return sample_batch

    @override(Policy)
    def get_weights(self) -> dict:
        return {
            "module": convert_to_non_torch_type(self.module.state_dict()),
            # Optimizer state dicts don't store tensors, only ids
            "optimizers": self.optimizers.state_dict(),
        }

    @override(Policy)
    def set_weights(self, weights: dict):
        self.module.load_state_dict(
            convert_to_torch_tensor(weights["module"], device=self.device)
        )
        # Optimizer state dicts don't store tensors, only ids
        self.optimizers.load_state_dict(weights["optimizers"])

    def convert_to_tensor(self, arr) -> Tensor:
        """Convert an array to a PyTorch tensor in this policy's device.

        Args:
            arr (array_like): object which can be converted using `np.asarray`
        """
        return convert_to_tensor(arr, self.device)

    def lazy_tensor_dict(self, sample_batch: SampleBatch) -> UsageTrackingDict:
        """Convert a sample batch into a dictionary of lazy tensors.

        The sample batch is wrapped with a UsageTrackingDict to convert array-
        likes into tensors upon querying.

        Args:
            sample_batch: the sample batch to convert

        Returns:
            A dictionary which intercepts key queries to lazily convert arrays
            to tensors.
        """
        tensor_batch = UsageTrackingDict(sample_batch)
        tensor_batch.set_get_interceptor(self.convert_to_tensor)
        return tensor_batch

    def __repr__(self):
        name = self.__class__.__name__
        args = [f"{self.observation_space},", f"{self.action_space},"]

        config = pretty_print(self.config).rstrip("\n")
        if "\n" in config:
            config = textwrap.indent(config, " " * 2)
            config = "{\n" + config + "\n}"

            args += [config]
            args_repr = "\n".join(args)
            args_repr = textwrap.indent(args_repr, " " * 2)
            constructor = f"{name}(\n{args_repr}\n)"
        else:
            args += [config]
            args_repr = " ".join(args[1:-1])
            constructor = f"{name}({args_repr})"
        return constructor

    def apply_gradients(self, gradients: ModelGradients) -> None:
        pass

    def compute_gradients(
        self, postprocessed_batch: SampleBatch
    ) -> Tuple[ModelGradients, Dict[str, TensorType]]:
        pass

    # ==========================================================================
    # InternalAPI
    # ==========================================================================

    def _build_config(self, config: dict) -> dict:
        if not self.options.all_options_set:
            raise RuntimeError(
                f"{type(self).__name__} still has configs to be set."
                " Did you call `configure` as the last decorator?"
            )

        passed = config.get("policy", {}).copy()
        passed.update({k: config[k] for k in self.pull_from_global if k in config})
        new = self.options.merge_defaults_with(passed)
        return new

    @staticmethod
    def _make_module(obs_space: Space, action_space: Space, config: dict) -> nn.Module:
        """Build the PyTorch nn.Module to be used by this policy.

        Args:
            obs_space: the observation space for this policy
            action_space: the action_space for this policy
            config: the user config containing the 'module' key

        Returns:
            A neural network module.
        """
        return get_module(obs_space, action_space, config["module"])

    def _make_optimizers(self) -> OptimizerCollection:
        """Build PyTorch optimizers to use.

        The result will be set as the policy's optimizer collection.

        The user should update the optimizer collection (mutable mapping)
        returned by the base implementation.

        Returns:
            A mapping from names to optimizer instances
        """
        # pylint:disable=no-self-use
        return OptimizerCollection()

    def _unpack_observations(self, input_dict):
        restored = input_dict.copy()
        restored["obs"] = restore_original_dimensions(
            input_dict["obs"], self.observation_space, self.framework
        )
        if len(input_dict["obs"].shape) > 2:
            restored["obs_flat"] = flatten(input_dict["obs"], self.framework)
        else:
            restored["obs_flat"] = input_dict["obs"]
        return restored

    def _compute_module_output(
        self, input_dict: TensorDict, state: List[Tensor], seq_lens: Tensor
    ) -> Tuple[TensorDict, List[Tensor]]:
        """Call the module with the given input tensors and state.

        This mirrors the method used by RLlib to execute the forward pass. Nested
        observation tensors are unpacked before this function is called.

        Subclasses should override this for custom forward passes (e.g., for recurrent
        networks).

        Args:
            input_dict: dictionary of input tensors, including "obs",
                "prev_action", "prev_reward", "is_training"
            state: list of state tensors with sizes matching those returned
                by get_initial_state + the batch dimension
            seq_lens: 1d tensor holding input sequence lengths

        Returns:
            A tuple containg an input dictionary to the policy's `dist_class`
            and a list of rnn state tensors
        """
        # pylint:disable=unused-argument,no-self-use
        return {"obs": input_dict["obs"]}, state

    def _extra_action_out(
        self,
        input_dict: TensorDict,
        state_batches: List[Tensor],
        module: nn.Module,
        action_dist: ActionDistribution,
    ) -> StatDict:
        """Returns dict of extra info to include in experience batch.

        Args:
            input_dict: Dict of model input tensors.
            state_batches: List of state tensors.
            model: Reference to the model.
            action_dist: Action dist object
                to get log-probs (e.g. for already sampled actions).
        """
        # pylint:disable=unused-argument,no-self-use
        return {}

    # ==========================================================================
    # Unimplemented Policy methods
    # ==========================================================================

    def export_model(self, export_dir):
        pass

    def export_checkpoint(self, export_dir):
        pass

    def import_model_from_h5(self, import_file):
        pass



"""Collection of PyTorch distributions and related utilities."""
from .diag_multivariate_normal import DiagMultivariateNormal
from .logistic import Logistic
from .squashed_distribution import SquashedDistribution

__all__ = [
    "DiagMultivariateNormal",
    "Logistic",
    "SquashedDistribution",
]



# pylint:disable=missing-module-docstring
import torch
import torch.nn as nn
from torch import Tensor


class TensorStandardScaler(nn.Module):
    # pylint:disable=line-too-long
    """Shifts and scales 1D inputs for model networks.

    Based on MBPO's `TensorStandardScaler`_

    .. _`TensorStandardScaler`: https://github.com/JannerM/mbpo/blob/28b1e3b1382dcda34b421961286b59f77770d48c/mbpo/models/utils.py#L14

    Args:
        input_size: Size of the inputs to the scaler.

    Notes:
        Loss functions for model ensembles already expand the inputs for the
        ensemble size, so the input size used here should correspond to the raw
        observation/action size.
    """
    # pylint:enable=line-too-long

    def __init__(self, input_size: int):
        super().__init__()
        self.register_buffer("scaler_mu", torch.zeros(input_size))
        self.register_buffer("scaler_sigma", torch.ones(input_size))
        self.fitted = False

    @torch.jit.export
    def fit(self, data: Tensor):
        """Assigns the data's mean and stddev to internal buffers.

        Runs two ops, one for assigning the mean of the data to the internal
        mean, and another for assigning the standard deviation of the data to
        the internal standard deviation.

        Arguments:
            data: Tensor containing the input
        """
        # Reduce all but the last dimension
        # pylint:disable=unnecessary-comprehension
        reduction_dims = [i for i in range(data.dim() - 1)]
        # pylint:enable=unnecessary-comprehension

        data_mu = torch.mean(data, dim=reduction_dims, keepdim=False)
        data_sigma = torch.std(data, dim=reduction_dims, keepdim=False)
        data_sigma = torch.where(
            data_sigma < 1e-12, torch.ones_like(data_sigma), data_sigma
        )

        self.scaler_mu.copy_(data_mu)
        self.scaler_sigma.copy_(data_sigma)
        self.fitted = True

    def forward(self, data: Tensor) -> Tensor:
        """Transforms the input matrix data using the parameters of this scaler.

        Args:
            data: Tensor containing the points to be transformed.

        Returns:
            The transformed dataset.
        """
        # pylint:disable=arguments-differ
        return (data - self.scaler_mu) / self.scaler_sigma

    @torch.jit.export
    def inverse_transform(self, data: Tensor) -> Tensor:
        """Undoes the transformation performed by this scaler.

        Args:
            data: Tensor containing the points to be transformed.

        Returns:
            The transformed dataset.
        """
        return self.scaler_sigma * data + self.scaler_mu



# pylint:disable=missing-module-docstring
import torch


def sum_rightmost(value: torch.Tensor, dim: int):
    """Sum out ``dim`` many rightmost dimensions of a given tensor.

    Args:
        value: A tensor of ``.dim()`` at least ``dim``.
        dim: The number of rightmost dims to sum out.
    """
    if dim == 0:
        return value
    required_shape = value.shape[:-dim] + (-1,)
    return value.reshape(required_shape).sum(-1)


def multiply_rightmost(value: torch.Tensor, dim: int):
    """Multiply out ``dim`` many rightmost dimensions of a given tensor.

    Args:
        value: A tensor of ``.dim()`` at least ``dim``.
        dim: The number of rightmost dims to multiply out.
    """
    if dim == 0:
        return value
    required_shape = value.shape[:-dim] + (-1,)
    return value.reshape(required_shape).prod(-1)



# pylint:disable=missing-module-docstring
from typing import Dict, Tuple

from torch import Tensor

__all__ = ["DistParams", "SampleLogp"]

DistParams = Dict[str, Tensor]
SampleLogp = Tuple[Tensor, Tensor]



"""Network and configurations for modules with stochastic policies."""
import warnings
from dataclasses import dataclass, field
from typing import Union

import torch.nn as nn
from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box, Discrete

from .policy.stochastic import (
    Alpha,
    MLPContinuousPolicy,
    MLPDiscretePolicy,
    MLPStochasticPolicy,
)

MLPSpec = MLPStochasticPolicy.spec_cls


@dataclass
class StochasticActorSpec(DataClassJsonMixin):
    """Specifications for stochastic policy and entropy coefficient.

    Args:
        encoder: Specifications for building the multilayer perceptron state
            processor
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state
        initial_entropy_coeff: Optional initial value of the entropy bonus term.
            The actor creates an `alpha` attribute with this initial value.
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments.
    """

    encoder: MLPSpec = field(default_factory=MLPSpec)
    input_dependent_scale: bool = False
    initial_entropy_coeff: float = 1.0
    initializer: dict = field(default_factory=dict)

    def __post_init__(self):
        cls_name = type(self).__name__
        ent_coeff = self.initial_entropy_coeff
        if ent_coeff < 0:
            warnings.warn(f"Entropy coefficient is negative in {cls_name}: {ent_coeff}")


class StochasticActor(nn.Module):
    """NN with stochastic policy.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for stochastic policy

    Attributes:
        policy: Stochastic policy to be learned
        alpha: Entropy bonus coefficient
    """

    # pylint:disable=abstract-method
    spec_cls = StochasticActorSpec

    def __init__(
        self,
        obs_space: Box,
        action_space: Union[Box, Discrete],
        spec: StochasticActorSpec,
    ):
        super().__init__()

        if isinstance(action_space, Box):
            policy = MLPContinuousPolicy(
                obs_space, action_space, spec.encoder, spec.input_dependent_scale
            )
        elif isinstance(action_space, Discrete):
            policy = MLPDiscretePolicy(obs_space, action_space, spec.encoder)
        else:
            raise ValueError(f"Unsopported action space type {type(action_space)}")
        policy.initialize_parameters(spec.initializer)

        self.policy = policy
        self.alpha = Alpha(spec.initial_entropy_coeff)



# pylint:disable=missing-docstring
import torch
import torch.nn as nn
from ray.rllib.utils import override


class LeafParameter(nn.Module):
    """Holds a single paramater vector an expands it to match batch shape of inputs."""

    def __init__(self, in_features):
        super().__init__()
        self.bias = nn.Parameter(torch.zeros(in_features))

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.bias.expand(inputs.shape[:-1] + (-1,))



"""Common type annotations."""
from typing import Tuple

from torch import Tensor

LQR = Tuple[Tensor, Tensor, Tensor, Tensor]
Affine = Tuple[Tensor, Tensor]
Quadratic = Tuple[Tensor, Tensor]
Box = Tuple[Tensor, Tensor]



"""Utilities for building optimizers."""
import contextlib
from typing import Type

import torch
import torch.nn as nn
from torch.optim import Optimizer

from raylab.utils.dictionaries import all_except

from .kfac import EKFAC, KFAC
from .radam import AdamW, PlainRAdam, RAdam

OPTIMIZERS = {
    name: cls
    for name, cls in [(k, getattr(torch.optim, k)) for k in dir(torch.optim)]
    if isinstance(cls, type) and issubclass(cls, Optimizer) and cls is not Optimizer
}
OPTIMIZERS.update(
    {
        "KFAC": KFAC,
        "EKFAC": EKFAC,
        "RAdam": RAdam,
        "PlainRAdam": PlainRAdam,
        "AdamW": AdamW,
    }
)


def build_optimizer(module: nn.Module, config: dict, wrap: bool = False) -> Optimizer:
    """Build optimizer with the desired config and tied to a module.

    Args:
        module: the module to tie the optimizer to (or its parameters)
        config: mapping containing the 'type' of the optimizer and additional
            kwargs.
        wrap: whether to wrap the class with :func:`wrap_optim_cls`.
    """
    cls = get_optimizer_class(config["type"], wrap=wrap)
    return link_optimizer(cls, module, all_except(config, "type"))


def get_optimizer_class(name: str, wrap: bool = True) -> Type[Optimizer]:
    """Return the optimizer class given its name.

    Args:
        name: the optimizer's name
        wrap: whether to wrap the class with :func:`wrap_optim_cls`.

    Returns:
        The optimizer class
    """
    try:
        cls = OPTIMIZERS[name]
        return wrap_optim_cls(cls) if wrap else cls
    except KeyError:
        # pylint:disable=raise-missing-from
        raise ValueError(f"Couldn't find optimizer with name '{name}'")


def wrap_optim_cls(cls: Type[Optimizer]) -> Type[Optimizer]:
    """Return PyTorch optimizer with additional context manager."""

    class ContextManagerOptim(cls):  # type:ignore
        # pylint:disable=missing-class-docstring,too-few-public-methods
        @contextlib.contextmanager
        def optimize(self):
            """Zero grads before yielding and step the optimizer upon exit."""
            self.zero_grad()
            yield
            self.step()

    return ContextManagerOptim


def link_optimizer(cls: Type[Optimizer], module: nn.Module, config: dict) -> Optimizer:
    """Construct optimizer tied to a module.

    Args:
        cls: the type of the optimizer
        module: the neural network module
        config: options to pass to the optimizer's constructor

    Returns:
        The optimizer instance
    """
    if issubclass(cls, (EKFAC, KFAC)):
        return cls(module, **config)
    return cls(module.parameters(), **config)



"""
Linear Quadratic Regulator (LQR):
Please see http://rail.eecs.berkeley.edu/deeprlcourse/static/slides/lec-10.pdf
for notation and more details on LQR.
"""
from typing import List

import torch
import torch.nn as nn
from torch import Tensor

from .types import LQR, Affine


class LQRSim(nn.Module):
    """Linear Quadratic Regulator simulator."""

    # pylint:disable=invalid-name,abstract-method,missing-function-docstring
    def __init__(self, system: LQR):
        super().__init__()
        F, f, C, c = system
        self.F = F.float().detach()
        self.f = f.float().detach()
        self.C = C.float().detach()
        self.c = c.float().detach()

    @torch.jit.export
    def transition(self, x, u):
        inputs = torch.cat([x, u])
        return self.F @ inputs + self.f

    @torch.jit.export
    def cost(self, x, u):
        inputs = torch.cat([x, u])
        c1 = 1 / 2 * inputs.T @ self.C @ inputs
        c2 = inputs @ self.c
        return c1 + c2

    @torch.jit.export
    def final_cost(self, x):
        state_size = self.state_size
        C_xx = self.C[:state_size, :state_size]
        c_x = self.c[:state_size]
        c1 = 1 / 2 * x.T @ C_xx @ x
        c2 = x @ c_x
        return c1 + c2

    @torch.jit.export
    def forward(self, policy: List[Affine], x0: Tensor):
        # pylint:disable=arguments-differ
        states = [x0]
        actions = []
        costs = []

        state = x0

        for K, k in policy:
            action = K @ state + k

            next_state = self.transition(state, action)
            cost = self.cost(state, action)

            state = next_state

            states.append(next_state)
            actions.append(action)
            costs.append(cost)

        final_cost = self.final_cost(state)
        costs.append(final_cost)

        return torch.stack(states), torch.stack(actions), torch.stack(costs)



"""Utilities for incorporating Weights & Biases with Raylab trainers."""
import numbers

from ray.tune.utils import flatten_dict

try:
    import wandb
except ImportError:
    wandb = None


class WandBLogger:
    """Synchronizes training results with Weights&Biases.

    Special configurations include:
    * file_paths: Sequence of file names to pass to `wandb.save` on trainer
      setup.
    * save_checkpoints: Whether to sync trainer checkpoints to W&B via
      `wandb.save`.
    """

    # pylint:disable=missing-docstring

    SPECIAL_KEYS = {"file_paths", "save_checkpoints"}

    def __init__(self, config: dict, name: str):
        create = bool(config.get("wandb", {}))
        self._run = self._setup(config, name) if create else None

    def _setup(self, config: dict, name: str):
        assert wandb is not None, "Unable to import wandb, did you install it via pip?"

        wandb_config = config["wandb"]
        self._save_checkpoints = wandb_config.get("save_checkpoints", False)

        config_exclude_keys = {"wandb", "callbacks"}
        config_exclude_keys.update(set(wandb_config.pop("config_exclude_keys", [])))

        wandb_kwargs = dict(
            name=name,
            config_exclude_keys=config_exclude_keys,
            config=config,
            # Allow calling init twice if creating more than one trainer in the
            # same process
            reinit=True,
        )
        wandb_kwargs.update(
            {k: v for k, v in wandb_config.items() if k not in self.SPECIAL_KEYS}
        )
        run = wandb.init(**wandb_kwargs)

        file_paths = wandb_config.get("file_paths", [])
        for path in file_paths:
            wandb.save(path)

        return run

    @property
    def enabled(self):
        return self._run is not None

    @property
    def run(self):
        return self._run

    @staticmethod
    def log_result(result: dict):
        # Avoid logging the config every iteration
        # Only log Jsonable objects
        tmp = result.copy()
        for k in ["done", "config", "pid", "timestamp"]:
            if k in tmp:
                del tmp[k]
        metrics = {}
        for key, value in flatten_dict(tmp, delimiter="/").items():
            if not isinstance(value, numbers.Number):
                continue
            metrics[key] = value
        wandb.log(metrics)

    def save_checkpoint(self, checkpoint_path: str):
        if self._save_checkpoints:
            wandb.save(checkpoint_path)

    @staticmethod
    def stop():
        wandb.join()



# pylint:disable=missing-module-docstring
from typing import Dict, List

import torch
import torch.nn as nn
from ray.rllib.utils import override

from .abstract import ConditionalDistribution


class Uniform(ConditionalDistribution):
    """
    Generates uniformly distributed random samples from the half-open interval
    ``[low, high)``.
    """

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ,no-self-use
        low, high = torch.chunk(inputs, 2, dim=-1)
        return {"low": low, "high": high}

    @override(ConditionalDistribution)
    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out = self._gen_sample(params, sample_shape).detach()
        return out, self.log_prob(out, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def rsample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        out = self._gen_sample(params, sample_shape)
        return out, self.log_prob(out, params)

    def _gen_sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int]):
        low, high = self._unpack_params(params)
        shape = sample_shape + low.shape
        rand = torch.rand(shape, dtype=low.dtype, device=low.device)
        return low + rand * (high - low)

    @override(ConditionalDistribution)
    @torch.jit.export
    def log_prob(self, value, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        lbound = low.le(value).type_as(low)
        ubound = high.gt(value).type_as(low)
        return torch.log(lbound.mul(ubound)) - torch.log(high - low)

    @override(ConditionalDistribution)
    @torch.jit.export
    def cdf(self, value, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        result = (value - low) / (high - low)
        return result.clamp(min=0, max=1)

    @override(ConditionalDistribution)
    @torch.jit.export
    def icdf(self, value, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        return value * (high - low) + low

    @override(ConditionalDistribution)
    @torch.jit.export
    def entropy(self, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        return torch.log(high - low)

    @override(ConditionalDistribution)
    @torch.jit.export
    def reproduce(self, value, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        rand = (value - low) / (high - low)
        sample_ = low + rand.detach() * (high - low)
        return sample_, self.log_prob(sample_, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        low, high = self._unpack_params(params)
        sample = (low + high) / 2
        return sample, self.log_prob(sample, params)

    def _unpack_params(self, params: Dict[str, torch.Tensor]):
        # pylint:disable=no-self-use
        return params["low"], params["high"]



# pylint:disable=missing-module-docstring
from dataclasses import dataclass

import torch
import torch.nn as nn
from gym.spaces import Box
from ray.rllib.utils import override

import raylab.torch.nn as nnx
import raylab.torch.nn.distributions as ptd
from raylab.policy.modules.networks.mlp import StateActionMLP
from raylab.torch.nn.init import initialize_

from .single import StochasticModel


@dataclass
class SVGModelSpec(StateActionMLP.spec_cls):
    """Specifications for stochastic mlp model network.

    Inherits parameters from `StateActionMLP.spec_cls`.

    Args:
        units: Number of units in each hidden layer
        activation: Nonlinearity following each linear layer
        delay_action: Whether to apply an initial preprocessing layer on the
            observation before concatenating the action to the input.
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state and action
        residual: Whether to build model as a residual one, i.e., that
            predicts the change in state rather than the next state itself
    """

    input_dependent_scale: bool = True
    residual: bool = True


class SVGModel(StochasticModel):
    """Model from Stochastic Value Gradients."""

    # pylint:disable=abstract-method
    spec_cls = SVGModelSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: SVGModelSpec):
        params = SVGDynamicsParams(obs_space, action_space, spec)
        dist = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)
        super().__init__(params, dist)

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all encoder parameters.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.params.initialize_parameters(initializer_spec)


class SVGDynamicsParams(nn.Module):
    """
    Neural network module mapping inputs to distribution parameters through parallel
    subnetworks for each output dimension.
    """

    # pylint:disable=abstract-method
    spec_cls = SVGModelSpec

    def __init__(self, obs_space, action_space, spec: SVGModelSpec):
        super().__init__()

        def make_encoder():
            return StateActionMLP(obs_space, action_space, spec)

        self.logits = nn.ModuleList([make_encoder() for _ in range(obs_space.shape[0])])
        self._activation = spec.activation

        def make_param(in_features):
            kwargs = dict(
                event_size=1, input_dependent_scale=spec.input_dependent_scale
            )
            return nnx.NormalParams(in_features, **kwargs)

        self.params = nn.ModuleList([make_param(m.out_features) for m in self.logits])

    @override(nn.Module)
    def forward(self, obs, act):  # pylint:disable=arguments-differ
        params = [p(l(obs, act)) for p, l in zip(self.params, self.logits)]
        loc = torch.cat([d["loc"] for d in params], dim=-1)
        scale = torch.cat([d["scale"] for d in params], dim=-1)
        return {"loc": loc, "scale": scale}

    def initialize_parameters(self, initializer_spec: dict):
        # pylint:disable=missing-docstring
        self.logits.apply(initialize_(activation=self._activation, **initializer_spec))



"""Custom activation functions as neural network modules."""
import torch
import torch.nn as nn


class Swish(nn.Module):
    r"""Swish activation function.

    Notes:
        Applies the mapping :math:`x \mapsto x \cdot \sigma(x)`,
        where :math:`sigma` is the sigmoid function.

    Reference:
        Eger, Steffen, Paul Youssef, and Iryna Gurevych.
        "Is it time to swish? Comparing deep learning activation functions
        across NLP tasks."
        arXiv preprint arXiv:1901.02671 (2019).
    """
    # pylint:disable=arguments-differ,no-self-use

    def forward(self, value: torch.Tensor) -> torch.Tensor:
        return value * value.sigmoid()



# pylint:disable=missing-docstring
import torch
import torch.nn as nn
from ray.rllib.utils import override


class TrilMatrix(nn.Module):
    """Neural network module which outputs a lower-triangular matrix."""

    __constants__ = {"in_features", "matrix_dim", "row_sizes"}

    def __init__(self, in_features, matrix_dim):
        super().__init__()
        self.in_features = in_features
        self.matrix_dim = matrix_dim
        self.row_sizes = tuple(range(1, self.matrix_dim + 1))
        tril_dim = int(self.matrix_dim * (self.matrix_dim + 1) / 2)
        self.linear_module = nn.Linear(self.in_features, tril_dim)

    @override(nn.Module)
    def forward(self, logits):  # pylint:disable=arguments-differ
        # Batch of flattened lower triangular matrices: [..., N * (N + 1) / 2]
        flat_tril = self.linear_module(logits)
        # Split flat lower triangular into rows
        split_tril = torch.split(flat_tril, self.row_sizes, dim=-1)
        # Compute exponentiated diagonals, row by row
        tril_rows = []
        for row in split_tril:
            zeros = torch.zeros(row.shape[:-1] + (self.matrix_dim - row.shape[-1],))
            decomposed_row = [row[..., :-1], row[..., -1:] ** 2, zeros]
            tril_rows.append(torch.cat(decomposed_row, dim=-1))
        # Stack rows into a single (batched) matrix. dim=-2 ensures that we stack then
        # as rows, not columns (which would effectively transpose the matrix into an
        # upper triangular one)
        tril = torch.stack(tril_rows, dim=-2)
        return tril



# pylint:disable=missing-docstring
# pylint: enable=missing-docstring
import torch
import torch.distributions as ptd


class Logistic(ptd.TransformedDistribution):
    """Creates a Logistic distribution parameterized by loc and scale."""

    # pylint:disable=abstract-method

    def __init__(self, loc, scale, **kwargs):
        loc, scale = map(torch.as_tensor, (loc, scale))
        base_distribution = ptd.Uniform(
            torch.zeros_like(loc), torch.ones_like(loc), **kwargs
        )
        transforms = [
            ptd.SigmoidTransform().inv,
            ptd.AffineTransform(loc=loc, scale=scale),
        ]
        super().__init__(base_distribution, transforms)



"""
Linear Quadratic Regulator (LQR):
Please see http://rail.eecs.berkeley.edu/deeprlcourse/static/slides/lec-10.pdf
for notation and more details on LQR.
"""
from typing import List, Tuple

import torch
import torch.nn as nn
from torch import Tensor

from .types import LQR as System
from .types import Affine, Quadratic

Policy = Affine
Value = Tuple[Tensor, Tensor, Tensor]


class LQRSolver(nn.Module):
    """Linear Quadratic Regulator solver."""

    # pylint:disable=abstract-method,invalid-name,missing-function-docstring,no-self-use

    def forward(self, LQR: System, T: int):
        # pylint:disable=too-many-arguments,arguments-differ
        policy: List[Policy] = []
        value_fn: List[Value] = []

        Vs = self.final_V(LQR)
        value_fn.append(Vs)

        for _ in range(T):  # Effectively solving backwards through time
            Ks, Vs = self.single_step(LQR, Vs)
            policy.insert(0, Ks)
            value_fn.insert(0, Vs)

        return policy, value_fn[:-1]

    def final_V(self, LQR: System) -> Value:
        _, _, C, c = LQR
        state_size = self.state_size(LQR)

        Vs = (C[:state_size, :state_size], c[:state_size], torch.zeros(()))
        return Vs

    def single_step(self, LQR: System, Vs: Value):
        Qs = self.compute_Q(LQR, Vs)
        Ks = self.compute_K(LQR, Qs)
        Vs = self.compute_V(LQR, Qs, Ks, Vs)
        return Ks, Vs

    def compute_Q(self, LQR: System, Vs: Value):
        F, f, C, c = LQR
        V, v, _ = Vs

        FV = F.T @ V
        Q = C + FV @ F
        q = c + FV @ f + F.T @ v
        return Q, q

    def compute_K(self, LQR: System, Qs: Quadratic):
        state_size = self.state_size(LQR)
        Q, q = Qs
        Q_uu = Q[state_size:, state_size:]
        Q_ux = Q[state_size:, :state_size]
        q_u = q[state_size:]

        inv_Q_uu = Q_uu.inverse()

        K = -inv_Q_uu @ Q_ux
        k = -inv_Q_uu @ q_u
        return K, k

    def compute_V(self, LQR: System, Qs: Quadratic, Ks: Affine, Vs: Value):
        # pylint:disable=too-many-locals
        state_size = self.state_size(LQR)
        Q, q = Qs
        Q_uu = Q[state_size:, state_size:]
        Q_ux = Q[state_size:, :state_size]
        q_u = q[state_size:]
        Q_xx = Q[:state_size, :state_size]
        Q_xu = Q[:state_size, state_size:]
        q_x = q[:state_size]

        K, k = Ks
        K_Q_uu = K.T @ Q_uu

        V = Q_xx + Q_xu @ K + K.T @ Q_ux + K_Q_uu @ K
        v = q_x + Q_xu @ k + K.T @ q_u + K_Q_uu @ k
        const = self.compute_const(LQR, Vs, Ks)
        return V, v, const

    def compute_const(self, LQR: System, Vs: Value, Ks: Affine):
        # pylint:disable=too-many-locals
        W_uu, w_u, V_f = self.compute_W(LQR, Vs)
        _, f, _, _ = LQR
        _, v, const_0 = Vs
        _, k = Ks
        const1 = 1 / 2 * k.T @ W_uu @ k
        const2 = k.T @ w_u
        const3 = 1 / 2 * f.T @ V_f + f.T @ v
        const = const_0 + const1 + const2 + const3
        return const.squeeze()

    def compute_W(self, LQR: System, Vs: Value):
        # pylint:disable=too-many-locals
        state_size = self.state_size(LQR)
        F, f, C, c = LQR

        V, v, _ = Vs
        V_f = V @ f
        W = C + F.T @ V @ F
        w = c + F.T @ V_f + F.T @ v
        W_uu = W[state_size:, state_size:]
        w_u = w[state_size:]

        return W_uu, w_u, V_f

    def state_size(self, LQR: System):
        F, _, _, _ = LQR
        return F.shape[0]



"""Modules mapping inputs to distribution parameters."""
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from ray.rllib.utils import override

from raylab.torch.nn.init import initialize_

from .leaf_parameter import LeafParameter


class CategoricalParams(nn.Module):
    """Produce Categorical parameters.

    Initialized to be close to a discrete uniform distribution.
    """

    def __init__(self, in_features: int, n_categories: int):
        super().__init__()
        self.logits_module = nn.Linear(in_features, n_categories)
        self.apply(initialize_("orthogonal", gain=0.01))

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        logits = self.logits_module(inputs)
        return {"logits": logits - logits.logsumexp(dim=-1, keepdim=True)}


class PolicyNormalParams(nn.Module):
    # pylint:disable=line-too-long
    """Produce Normal parameters for a stochastic policy.

    Max and min log scale clipping borrowed from `spinningup`_ and `pfrl`_.
    Initialization copied from `pfrl`_.

    .. _`spinningup`: https://github.com/openai/spinningup/blob/master/spinup/algos/pytorch/sac/core.py
    .. _`prfrl`: https://github.com/pfnet/pfrl/blob/master/examples/mujoco/reproduction/soft_actor_critic/train_soft_actor_critic.py
    """
    # pylint:enable=line-too-long
    def __init__(
        self,
        in_features: int,
        event_size: int,
        input_dependent_scale: bool = True,
    ):
        super().__init__()
        self.loc_module = nn.Linear(in_features, event_size)
        if input_dependent_scale:
            self.log_scale_module = nn.Linear(in_features, event_size)
        else:
            self.log_scale_module = LeafParameter(event_size)

        self.apply(initialize_("xavier_uniform", gain=1.0))

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        loc = self.loc_module(inputs)

        log_scale = self.log_scale_module(inputs)
        log_scale = torch.clamp(log_scale, min=-20, max=2)
        scale = log_scale.exp()

        return {"loc": loc, "scale": scale}


class NormalParams(nn.Module):
    # pylint:disable=line-too-long
    """Produce Normal parameters.

    Initialized to be close to a standard Normal distribution.

    Utilizes bounded log_stddev as described in the 'Well behaved probabilistic
    networks' appendix of `PETS`_.

    .. _`PETS`: https://papers.nips.cc/paper/7725-deep-reinforcement-learning-in-a-handful-of-trials-using-probabilistic-dynamics-models
    """
    # pylint:enable=line-too-long

    def __init__(
        self,
        in_features: int,
        event_size: int,
        input_dependent_scale: bool = True,
        bound_parameters: bool = True,
    ):
        super().__init__()
        self.loc_module = nn.Linear(in_features, event_size)
        if input_dependent_scale:
            self.log_scale_module = nn.Linear(in_features, event_size)
        else:
            self.log_scale_module = LeafParameter(event_size)

        max_logvar = torch.ones(event_size) / 2.0
        min_logvar = -torch.ones(event_size) * 10.0
        if bound_parameters:
            self.max_logvar = nn.Parameter(max_logvar)
            self.min_logvar = nn.Parameter(min_logvar)
        else:
            self.register_buffer("max_logvar", max_logvar)
            self.register_buffer("min_logvar", min_logvar)

        self.apply(initialize_("orthogonal", gain=0.01))

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        loc = self.loc_module(inputs)

        log_scale = self.log_scale_module(inputs)
        max_logvar = self.max_logvar.expand_as(log_scale)
        min_logvar = self.min_logvar.expand_as(log_scale)
        log_scale = max_logvar - F.softplus(max_logvar - log_scale)
        log_scale = min_logvar + F.softplus(log_scale - min_logvar)
        scale = log_scale.exp()

        return {
            "loc": loc,
            "scale": scale,
            "max_logvar": max_logvar,
            "min_logvar": min_logvar,
        }


class StdNormalParams(nn.Module):
    """Produce Normal parameters with unit variance."""

    def __init__(self, input_dim: int, event_size: int):
        super().__init__()
        self.input_dim = input_dim
        self.event_shape = (event_size,)

    @override(nn.Module)
    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        # pylint:disable=arguments-differ
        shape = inputs.shape[: -self.input_dim] + self.event_shape
        return {"loc": torch.zeros(shape), "scale": torch.ones(shape)}



"""NN modules for stochastic dynamics estimation."""
from dataclasses import dataclass
from typing import List

import torch
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor

import raylab.torch.nn as nnx
import raylab.torch.nn.distributions as ptd
from raylab.policy.modules.networks.mlp import StateActionMLP
from raylab.torch.nn.distributions.types import SampleLogp
from raylab.utils.types import TensorDict


class StochasticModel(nn.Module):
    """Represents a stochastic model as a conditional distribution module."""

    # pylint:disable=abstract-method
    def __init__(
        self, params_module: nn.Module, dist_module: ptd.ConditionalDistribution
    ):
        super().__init__()
        self.params = params_module
        self.dist = dist_module

    def forward(self, obs, action) -> TensorDict:  # pylint:disable=arguments-differ
        """Compute state-action conditional distribution parameters."""
        return self.params(obs, action)

    @torch.jit.export
    def sample(self, params: TensorDict, sample_shape: List[int] = ()) -> SampleLogp:
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        return self.dist.sample(params, sample_shape)

    @torch.jit.export
    def rsample(self, params: TensorDict, sample_shape: List[int] = ()) -> SampleLogp:
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        return self.dist.rsample(params, sample_shape)

    @torch.jit.export
    def log_prob(self, next_obs: Tensor, params: TensorDict) -> Tensor:
        """
        Returns the log probability density/mass function evaluated at `next_obs`.
        """
        return self.dist.log_prob(next_obs, params)

    @torch.jit.export
    def cdf(self, next_obs: Tensor, params: TensorDict) -> Tensor:
        """Returns the cumulative density/mass function evaluated at `next_obs`."""
        return self.dist.cdf(next_obs, params)

    @torch.jit.export
    def icdf(self, prob, params: TensorDict) -> Tensor:
        """Returns the inverse cumulative density/mass function evaluated at `prob`."""
        return self.dist.icdf(prob, params)

    @torch.jit.export
    def entropy(self, params: TensorDict) -> Tensor:
        """Returns entropy of distribution."""
        return self.dist.entropy(params)

    @torch.jit.export
    def perplexity(self, params: TensorDict) -> Tensor:
        """Returns perplexity of distribution."""
        return self.dist.perplexity(params)

    @torch.jit.export
    def reproduce(self, next_obs, params: TensorDict) -> SampleLogp:
        """Produce a reparametrized sample with the same value as `next_obs`."""
        return self.dist.reproduce(next_obs, params)

    @torch.jit.export
    def deterministic(self, params: TensorDict) -> SampleLogp:
        """
        Generates a deterministic sample or batch of samples if the distribution
        parameters are batched. Returns a (rsample, log_prob) pair.
        """
        return self.dist.deterministic(params)


class ResidualStochasticModel(StochasticModel):
    """Overrides StochasticModel interface to model state transition residuals."""

    # pylint:disable=abstract-method

    def __init__(self, model: StochasticModel):
        super().__init__(params_module=model.params, dist_module=model.dist)

    def forward(self, obs: Tensor, action: Tensor) -> TensorDict:
        params = self.params(obs, action)
        params["obs"] = obs
        return params

    @torch.jit.export
    def sample(self, params: TensorDict, sample_shape: List[int] = ()) -> SampleLogp:
        res, log_prob = self.dist.sample(params, sample_shape)
        return params["obs"] + res, log_prob

    @torch.jit.export
    def rsample(self, params: TensorDict, sample_shape: List[int] = ()) -> SampleLogp:
        res, log_prob = self.dist.rsample(params, sample_shape)
        return params["obs"] + res, log_prob

    @torch.jit.export
    def log_prob(self, next_obs: Tensor, params: TensorDict) -> Tensor:
        return self.dist.log_prob(next_obs - params["obs"], params)

    @torch.jit.export
    def cdf(self, next_obs: Tensor, params: TensorDict) -> Tensor:
        return self.dist.cdf(next_obs - params["obs"], params)

    @torch.jit.export
    def icdf(self, prob, params: TensorDict) -> Tensor:
        residual = self.dist.icdf(prob, params)
        return params["obs"] + residual

    @torch.jit.export
    def reproduce(self, next_obs, params: TensorDict) -> SampleLogp:
        sample_, log_prob_ = self.dist.reproduce(next_obs - params["obs"], params)
        return params["obs"] + sample_, log_prob_

    @torch.jit.export
    def deterministic(self, params: TensorDict) -> SampleLogp:
        sample, log_prob = self.dist.deterministic(params)
        return params["obs"] + sample, log_prob


class DynamicsParams(nn.Module):
    """Neural network mapping state-action pairs to distribution parameters.

    Args:
        encoder: Module mapping state-action pairs to 1D features
        params: Module mapping 1D features to distribution parameters
    """

    # pylint:disable=abstract-method
    def __init__(self, encoder: nn.Module, params: nn.Module):
        super().__init__()
        self.encoder = encoder
        self.params = params

    def forward(self, obs, actions):  # pylint:disable=arguments-differ
        """Compute state-action conditional distribution parameters."""
        return self.params(self.encoder(obs, actions))


@dataclass
class MLPModelSpec(StateActionMLP.spec_cls):
    """Specifications for stochastic mlp model network.

    Inherits parameters from `StateActionMLP.spec_cls`.

    Args:
        units: Number of units in each hidden layer
        activation: Nonlinearity following each linear layer
        delay_action: Whether to apply an initial preprocessing layer on the
            observation before concatenating the action to the input.
        fix_logvar_bounds: Whether to use fixed or dynamically adjusted
            bounds for the log-scale outputs of the network.
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state and action
    """

    fix_logvar_bounds: bool = True
    input_dependent_scale: bool = True


class MLPModel(StochasticModel):
    """Stochastic model with multilayer perceptron state-action encoder.

    Attributes:
        params: NN module mapping obs-act pairs to obs dist params
        dist: NN module implementing the distribution API
        encoder: NN module used in `params` to map obs-act pairs to vector
            embeddings
    """

    # pylint:disable=abstract-method
    spec_cls = MLPModelSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: MLPModelSpec):
        encoder = StateActionMLP(obs_space, action_space, spec)

        params = nnx.NormalParams(
            encoder.out_features,
            obs_space.shape[0],
            input_dependent_scale=spec.input_dependent_scale,
            bound_parameters=not spec.fix_logvar_bounds,
        )
        if spec.fix_logvar_bounds:
            params.max_logvar.fill_(2)
            params.min_logvar.fill_(-20)
        params = DynamicsParams(encoder, params)

        dist = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)

        super().__init__(params, dist)
        # Can only assign modules and parameters after calling nn.Module.__init__
        self.encoder = encoder

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all encoder parameters.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.encoder.initialize_parameters(initializer_spec)



"""Network and configurations for modules with Q-value critics."""
from dataclasses import dataclass, field

import torch.nn as nn
from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box

from .q_value import ForkedQValueEnsemble, MLPQValue, QValueEnsemble

QValueSpec = MLPQValue.spec_cls


@dataclass
class ActionValueCriticSpec(DataClassJsonMixin):
    """Specifications for action-value estimators.

    Args:
        encoder: Specifications for creating the multilayer perceptron mapping
            states and actions to pre-value function linear features
        double_q: Whether to create two Q-value estimators instead of one.
            Defaults to True
        parallelize: Whether to evaluate Q-values in parallel. Defaults to
            False.
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments.
    """

    encoder: QValueSpec = field(default_factory=QValueSpec)
    double_q: bool = True
    parallelize: bool = False
    initializer: dict = field(default_factory=dict)


class ActionValueCritic(nn.Module):
    """NN with Q-value estimators.

    Since it is common to use clipped double Q-Learning, `q_values` is a
    ModuleList of Q-value functions.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for action-value estimators

    Attributes:
        q_values: The action-value estimators to be learned
        target_q_values: The action-value estimators used for bootstrapping in
            Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = ActionValueCriticSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: ActionValueCriticSpec):
        super().__init__()

        def make_q_value():
            return MLPQValue(obs_space, action_space, spec.encoder)

        def make_q_value_ensemble():
            n_q_values = 2 if spec.double_q else 1
            q_values = [make_q_value() for _ in range(n_q_values)]

            if spec.parallelize:
                return ForkedQValueEnsemble(q_values)
            return QValueEnsemble(q_values)

        q_values = make_q_value_ensemble()
        q_values.initialize_parameters(spec.initializer)

        target_q_values = make_q_value_ensemble()
        main, target = set(q_values.parameters()), set(target_q_values.parameters())
        assert not main.intersection(
            target
        ), "Main and target Q nets cannot share params."
        target_q_values.load_state_dict(q_values.state_dict())
        for par in target:
            par.requires_grad = False

        self.q_values = q_values
        self.target_q_values = target_q_values



"""Implementations of stochastic dynamics models."""
from .builders import EnsembleSpec
from .builders import Spec as SingleSpec
from .builders import build as build_single
from .builders import build_ensemble
from .ensemble import SME, ForkedSME
from .single import MLPModel, ResidualStochasticModel, StochasticModel
from .svg import SVGModel



"""
MIT License

Copyright (c) 2018 Thomas George, César Laurent and Université de Montréal.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

Adapted from: https://github.com/Thrandis/EKFAC-pytorch
"""
from __future__ import annotations

import abc
import contextlib

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.optim import Optimizer


class KFACMixin(metaclass=abc.ABCMeta):
    """Adds methods for forward hooks, covariance computation and updating."""

    # pylint:disable=invalid-name,no-member
    _hookable_modules: list[nn.Module]

    @contextlib.contextmanager
    def record_stats(self):
        """Activate registered forward and backward hooks."""
        self.zero_grad()
        fwd_handles, bwd_handles = self.register_hooks()
        try:
            yield
        finally:
            self.remove_hooks(fwd_handles, bwd_handles)

    def register_hooks(self) -> tuple[list, list]:
        """Adds hooks to the monitored module for storing inputs and grads."""
        fwd_handles, bwd_handles = [], []
        for mod in self._hookable_modules:
            fwd_handles += [mod.register_forward_pre_hook(self.save_input)]
            bwd_handles += [mod.register_full_backward_hook(self.save_grad_out)]
        return fwd_handles, bwd_handles

    @staticmethod
    def remove_hooks(fwd_handles: list, bwd_handles: list):
        """Removes hooks from the monitored module."""
        for handle in fwd_handles + bwd_handles:
            handle.remove()

    def save_input(self, mod, inputs):
        """Saves input of layer to compute covariance.

        Note: inputs must be divided by the batch size to weight them appropriately
        when computing the average whitening matrix.
        """
        inputs = inputs[0].detach()
        if isinstance(mod, nn.Linear):
            inputs = inputs.reshape(-1, inputs.shape[-1])
        elif isinstance(mod, nn.Conv2d):
            inputs = inputs.reshape(-1, *inputs.shape[-3:])
        self.state[mod.weight]["x"] = inputs / inputs.shape[0]

    def save_grad_out(self, mod, _, grad_outputs):
        """Saves grad on output of layer to compute covariance.

        Note: grads are already properly weighted when the final loss function uses
        .mean() to aggregate element-wise losses. Since this is always the case when
        computing the entropy (average negative log-likelihood), we don't weight them
        here.
        """
        grad_outputs = grad_outputs[0].detach()
        if isinstance(mod, nn.Linear):
            grad_outputs = grad_outputs.reshape(-1, grad_outputs.shape[-1])
        elif isinstance(mod, nn.Conv2d):
            grad_outputs = grad_outputs.reshape(-1, *grad_outputs.shape[-3:])
        self.state[mod.weight]["gy"] = grad_outputs

    def step(self):  # pylint:disable=arguments-differ
        """Preconditions and applies gradients."""
        fisher_norm = 0.0
        for group in self.param_groups[:-1]:
            # Getting parameters
            params = group["params"]
            weight, bias = params if len(params) == 2 else (params[0], None)
            state = self.state[weight]

            # Update convariances and inverses
            state.setdefault("step", 0)
            state.update(self._compute_covs(group, state))
            if state["step"] % self.update_freq == 0:
                state.update(self._process_covs(state))
            state["step"] += 1

            # Preconditionning
            gw, gb, new_state = self._precond(weight, bias, group, state)
            state.update(new_state)
            # Updating gradients
            fisher_norm += (weight.grad * gw).sum()
            weight.grad.data = gw
            if bias is not None:
                fisher_norm += (bias.grad * gb).sum()
                bias.grad.data = gb

            # Cleaning
            self.state[weight].pop("x", None)
            self.state[weight].pop("gy", None)

        fisher_norm += sum(
            (p.grad * p.grad).sum() for p in self.param_groups[-1]["params"]
        )

        # Eventually scale the norm of the gradients and apply each
        scale = min(self.eta, torch.sqrt(self.state["kl_clip"] / fisher_norm))
        for group in self.param_groups:
            for param in group["params"]:
                param.grad.data.mul_(scale)
                param.data.sub_(param.grad.data, alpha=group["lr"])

    @abc.abstractmethod
    def _compute_covs(self, group, state):
        """Computes the covariances."""

    @abc.abstractmethod
    def _process_covs(self, state):
        """Process the covariances for preconditioning gradients later."""

    @abc.abstractmethod
    def _precond(self, weight, bias, group, state) -> tuple[Tensor, Tensor, dict]:
        """Applies preconditioning."""


class KFAC(KFACMixin, Optimizer):
    """K-FAC Optimizer for Linear and Conv2d layers.

    Computes the K-FAC of the second moment of the gradients.
    It works for Linear and Conv2d layers and silently skip other layers.

    Args:
        net: Network to optimize.
        eps: Tikhonov regularization parameter for the inverses.
        sua: Applies SUA (Spatially Uncorrelated Activations) approximation.
        pi: Computes pi correction for Tikhonov regularization.
        update_freq: Perform inverses every update_freq updates.
        alpha: Running average parameter (if == 1, no r. ave.).
        kl_clip: Scale the gradients by the squared fisher norm.
        eta: upper bound for gradient scaling.
    """

    # pylint:disable=invalid-name,too-many-instance-attributes

    def __init__(
        self,
        net: nn.Module,
        eps: float,
        sua: bool = False,
        pi: bool = False,
        update_freq: int = 1,
        alpha: float = 1.0,
        kl_clip: float = 1e-3,
        eta: float = 1.0,
        lr: float = 1.0,
    ):
        # pylint:disable=too-many-arguments,too-many-locals
        assert isinstance(net, nn.Module), "KFAC needs access to module structure."
        self.eps = eps
        self.sua = sua
        self.pi = pi
        self.update_freq = update_freq
        self.alpha = alpha
        self.eta = eta
        self._hookable_modules = []

        param_groups = []
        param_set = set()
        for mod in net.modules():
            mod_class = type(mod).__name__
            if mod_class in ["Linear", "Conv2d"]:
                self._hookable_modules += [mod]
                info = (
                    (mod.kernel_size, mod.padding, mod.stride)
                    if mod_class == "Conv2d"
                    else None
                )
                params = [mod.weight]
                if mod.bias is not None:
                    params.append(mod.bias)
                param_groups.append(
                    {"params": params, "info": info, "layer_type": mod_class}
                )
                param_set.update(set(params))

        param_groups.append(
            {"params": [p for p in net.parameters() if p not in param_set]}
        )
        super().__init__(param_groups, {"lr": lr})
        self.state["kl_clip"] = kl_clip

    def _compute_covs(self, group, state):
        x, gy = state["x"], state["gy"]
        # Computation of xxt
        if group["layer_type"] == "Conv2d":
            if not self.sua:
                kernel_size, padding, stride = group["info"]
                x = F.unfold(x, kernel_size, padding=padding, stride=stride)
            else:
                x = x.view(x.shape[0], x.shape[1], -1)
            x = x.data.permute(1, 0, 2).reshape(x.shape[1], -1)
        else:
            x = x.data.T

        if len(group["params"]) == 2:
            ones = torch.ones_like(x[:1])
            x = torch.cat([x, ones], dim=0)

        # Computation of xxt
        xxt = x @ x.T
        if "xxt" in state:
            xxt = state["xxt"] * (1.0 - self.alpha) + xxt * self.alpha

        # Computation of ggt
        if group["layer_type"] == "Conv2d":
            gy = gy.data.permute(1, 0, 2, 3)
            num_locations = gy.shape[2] * gy.shape[3]
            gy = gy.reshape(gy.shape[0], -1)
        else:
            gy = gy.data.T
            num_locations = 1

        ggt = gy @ gy.T
        if "ggt" in state:
            ggt = state["ggt"] * (1.0 - self.alpha) + ggt * self.alpha

        return {"xxt": xxt, "ggt": ggt, "num_locations": num_locations}

    def _process_covs(self, state):
        xxt, ggt, num_locations = (state[k] for k in "xxt ggt num_locations".split())
        # Computes pi
        pi = 1.0
        if self.pi:
            pi = (torch.trace(xxt) * ggt.shape[0]) / (torch.trace(ggt) * xxt.shape[0])

        # Regularizes and inverts
        eps = self.eps / num_locations
        diag_xxt = torch.diag(torch.empty(xxt.shape[0]).fill_(torch.sqrt(eps * pi)))
        diag_ggt = torch.diag(torch.empty(ggt.shape[0]).fill_(torch.sqrt(eps / pi)))
        ixxt = (xxt + diag_xxt).inverse()
        iggt = (ggt + diag_ggt).inverse()
        return {"ixxt": ixxt, "iggt": iggt}

    def _precond(self, weight, bias, group, state):
        if group["layer_type"] == "Conv2d" and self.sua:
            return self._precond_sua(weight, bias, state)

        g = weight.grad.data
        if group["layer_type"] == "Conv2d":
            g = g.reshape(g.shape[0], -1)
        if bias is not None:
            gb = bias.grad.data
            g = torch.cat([g, gb.view(gb.shape[0], 1)], dim=1)

        ixxt, iggt = state["ixxt"], state["iggt"]
        g = iggt @ g @ ixxt
        if group["layer_type"] == "Conv2d":
            g /= state["num_locations"]

        if bias is not None:
            gb = g[:, -1].reshape_as(bias)
            g = g[:, :-1]
        else:
            gb = None

        g = g.reshape_as(weight)
        return g, gb, {}

    @staticmethod
    def _precond_sua(weight, bias, state):
        """Preconditioning for KFAC SUA."""
        g = weight.grad.data
        s = g.shape
        g = g.permute(1, 0, 2, 3).contiguous()
        if bias is not None:
            gb = bias.grad.view(1, -1, 1, 1).expand(1, -1, s[2], s[3])
            g = torch.cat([g, gb], dim=0)

        ixxt, iggt = state["ixxt"], state["iggt"]
        g = ixxt @ g.reshape(-1, s[0] * s[2] * s[3])
        g = g.reshape(-1, s[0], s[2], s[3]).permute(1, 0, 2, 3)
        g = iggt @ g.reshape(s[0], -1)
        g = g.reshape(s[0], -1, s[2], s[3]) / state["num_locations"]
        if bias is not None:
            gb = g[:, -1, s[2] // 2, s[3] // 2]
            g = g[:, :-1]
        else:
            gb = None

        return g, gb, {}


class EKFAC(KFACMixin, Optimizer):
    """EKFAC Optimizer for Linear layers.

    It works for Linear layers and silently skip other layers.
    Note: unlike the paper's pseudocode, we maintain running averages of the
    Kronecker-factored covariance matrices.

    Args:
        net: Network to optimize.
        eps: Tikhonov regularization parameter for the inverses.
        update_freq: Perform inverses every update_freq updates.
        alpha: Running average parameter (if == 1, no r. ave.).
        kl_clip: Scale the gradients by the squared fisher norm.
        eta: upper bound for gradient scaling.
    """

    # pylint:disable=invalid-name

    def __init__(
        self,
        net: nn.Module,
        eps: float,
        update_freq: int = 1,
        alpha: float = 1.0,
        kl_clip: float = 1e-3,
        eta: float = 1.0,
        lr: float = 1.0,
    ):
        # pylint:disable=too-many-arguments
        assert isinstance(net, nn.Module), "EKFAC needs access to module structure."
        self.eps = eps
        self.update_freq = update_freq
        self.alpha = alpha
        self.eta = eta
        self._hookable_modules = []

        param_groups = []
        param_set = set()
        for mod in net.modules():
            mod_class = type(mod).__name__
            if mod_class in ["Linear"]:
                self._hookable_modules += [mod]
                info = None
                params = [mod.weight]
                if mod.bias is not None:
                    params.append(mod.bias)
                param_groups.append(
                    {"params": params, "info": info, "layer_type": mod_class}
                )
                param_set.update(set(params))

        param_groups.append(
            {"params": [p for p in net.parameters() if p not in param_set]}
        )
        super().__init__(param_groups, {"lr": lr})
        self.state["kl_clip"] = kl_clip

    def _compute_covs(self, group, state):
        x, gy = state["x"], state["gy"]

        # Computation of xxt
        x = x.data.T
        if len(group["params"]) == 2:
            x = torch.cat([x, torch.ones_like(x[:1])], dim=0)

        xxt = x @ x.T
        if "xxt" in state:
            xxt = state["xxt"] * (1.0 - self.alpha) + xxt * self.alpha

        # Computation of ggt
        gy = gy.data.T
        num_locations = 1

        ggt = gy @ gy.T
        if "ggt" in state:
            ggt = state["ggt"] * (1.0 - self.alpha) + ggt * self.alpha

        return {"xxt": xxt, "ggt": ggt, "num_locations": num_locations}

    def _process_covs(self, state):
        xxt, ggt = state["xxt"], state["ggt"]

        # Regularizes and inverts
        pi = (torch.trace(xxt) * ggt.shape[0]) / (torch.trace(ggt) * xxt.shape[0])
        eps = self.eps
        diag_xxt = torch.diag(torch.empty(xxt.shape[0]).fill_(torch.sqrt(eps * pi)))
        diag_ggt = torch.diag(torch.empty(ggt.shape[0]).fill_(torch.sqrt(eps / pi)))

        sa, ua = torch.symeig(xxt + diag_xxt, eigenvectors=True)
        sb, ub = torch.symeig(ggt + diag_ggt, eigenvectors=True)
        m2 = sb.unsqueeze(1) * sa.unsqueeze(0)
        return {"ua": ua, "ub": ub, "m2": m2}

    def _precond(self, weight, bias, group, state):
        g = weight.grad.data
        if bias is not None:
            gb = bias.grad.data
            g = torch.cat([g, gb.view(gb.shape[0], 1)], dim=1)
        bs = state["x"].size(0)

        ua, ub = state["ua"], state["ub"]
        projected = ub.T @ g @ ua

        m2 = projected ** 2
        if "m2" in state:
            m2 = state["m2"] * self.alpha + (1.0 - self.alpha) * bs * m2

        scaled = projected / (m2 + self.eps)
        g = ub @ scaled @ ua.T

        if bias is not None:
            gb = g[:, -1].reshape_as(bias)
            g = g[:, :-1]
        else:
            gb = None

        g = g.reshape_as(weight)
        return g, gb, {"m2": m2}



"""Network and configurations for modules with stochastic model ensembles."""
from typing import List

import torch
import torch.nn as nn
from torch import Tensor
from torch.jit import fork, wait

from raylab.torch.nn.distributions.types import SampleLogp
from raylab.utils.types import TensorDict

from .single import StochasticModel


class SME(nn.ModuleList):
    """Stochastic Model Ensemble.

    A static NN module list of `N` stochastic dynamics models. Implements the
    StochasticModel API but returns python lists of `N` outputs, one for each
    model in the ensemble.

    Assumes inputs are lists of the same length as the model ensemble.
    Applies each model in the ensemble to one of the inputs in the list.

    Args:
        models: List of StochasticModel modules

    Notes:
        `O` is the observation shape and `A` is the action shape.
    """

    # pylint:disable=abstract-method

    def __init__(self, models: List[StochasticModel]):
        cls_name = type(self).__name__
        assert all(
            isinstance(m, StochasticModel) for m in models
        ), f"All modules in {cls_name} must be instances of StochasticModel."
        super().__init__(models)

    def forward(self, obs: List[Tensor], act: List[Tensor]) -> List[TensorDict]:
        # pylint:disable=arguments-differ
        return [m(obs[i], act[i]) for i, m in enumerate(self)]

    @torch.jit.export
    def sample(self, params: List[TensorDict]) -> List[SampleLogp]:
        """Compute samples and likelihoods for each model in the ensemble.

        Args:
            obs: List of `N` observation tensors of shape `(*,) + O`
            action: List of `N` action tensors of shape `(*,) + A`
            sample_shape: Sample shape argument for each model in the ensemble

        Returns:
           List of `N` tuples of sample and log-likelihood tensors of shape
           `S + (*,) + O` and `S + (*,)` respectively, where `S` is the
           `sample_shape`.
        """
        return [m.sample(params[i]) for i, m in enumerate(self)]

    @torch.jit.export
    def rsample(self, params: List[TensorDict]) -> List[SampleLogp]:
        """Compute reparameterized samples and likelihoods for each model.

        Uses the same semantics as :meth:`SME.sample`.
        """
        return [m.rsample(params[i]) for i, m in enumerate(self)]

    @torch.jit.export
    def log_prob(self, new_obs: List[Tensor], params: List[TensorDict]) -> List[Tensor]:
        """Compute likelihoods for each model in the ensemble.

        Args:
            obs: List of `N` observation tensors of shape `(*,) + O`
            action: List of `N` action tensors of shape `(*,) + A`
            new_obs: List of `N` observation tensors of shape `(*,) + O`

        Returns:
           List of `N` log-likelihood tensors of shape `(*,)`
        """
        return [m.log_prob(new_obs[i], params[i]) for i, m in enumerate(self)]

    @torch.jit.export
    def deterministic(self, params: List[TensorDict]) -> List[SampleLogp]:
        """Compute deterministic new observations and their likelihoods for each model.

        Uses the same semantics as :meth:`SME.sample`.
        """
        return [m.deterministic(params[i]) for i, m in enumerate(self)]


class ForkedSME(SME):
    """Stochastic Model Ensemble with parallelized methods."""

    # pylint:disable=abstract-method

    def forward(self, obs: List[Tensor], act: List[Tensor]) -> List[TensorDict]:
        futures = [fork(m, obs[i], act[i]) for i, m in enumerate(self)]
        return [wait(f) for f in futures]

    @torch.jit.export
    def sample(self, params: List[TensorDict]) -> List[SampleLogp]:
        futures = [fork(m.sample, params[i]) for i, m in enumerate(self)]
        return [wait(f) for f in futures]

    @torch.jit.export
    def rsample(self, params: List[TensorDict]) -> List[SampleLogp]:
        futures = [fork(m.rsample, params[i]) for i, m in enumerate(self)]
        return [wait(f) for f in futures]

    @torch.jit.export
    def log_prob(self, new_obs: List[Tensor], params: List[TensorDict]) -> List[Tensor]:
        futures = [fork(m.log_prob, new_obs[i], params[i]) for i, m in enumerate(self)]
        return [wait(f) for f in futures]

    @torch.jit.export
    def deterministic(self, params: List[TensorDict]) -> List[SampleLogp]:
        futures = [fork(m.deterministic, params[i]) for i, m in enumerate(self)]
        return [wait(f) for f in futures]



"""
Hessian-free optimization utilities
"""
import torch
from torch.autograd import grad


def hessian_vector_product(output, params, vector):
    """Computes the Hessian vector product w.r.t to a scalar loss.

    Args:
        output (Tensor): loss tensor w.r.t. which the Hessian will be computed.
        params (list): the parameters of the module, usually from a call to
            `module.parameters()`.
        vector (Tensor): The flattened vector to compute the Hessian product with.
            This must have the same total number of elements in `params`.
    """
    # pylint:disable=missing-docstring
    params = list(params)
    vecs, idx = [], 0
    for par in params:
        vecs += [vector[idx : idx + par.numel()].reshape_as(par)]
        idx += par.numel()
    grads = grad(output, params, allow_unused=True, create_graph=True)
    hvp = grad([g for g in grads if g is not None], params, vecs, allow_unused=True)
    zeros = torch.zeros
    return torch.cat(
        [(zeros(p.numel()) if v is None else v.flatten()) for v, p in zip(hvp, params)]
    )


def conjugate_gradient(f_mat_vec_prod, b, cg_iters=10, residual_tol=1e-6):
    """
    Demmel p 312. Approximately solve x = A^{-1}b, or Ax = b,
    where we only have access to f: x -> Ax
    """
    # pylint:disable=invalid-name
    p = b.clone()
    r = b.clone()
    x = torch.zeros_like(b)
    rdotr = torch.dot(r, r)

    for i in range(cg_iters):
        z = f_mat_vec_prod(p)
        v = rdotr / torch.dot(p, z)
        x += v * p
        r -= v * z
        newrdotr = torch.dot(r, r)
        mu = newrdotr / rdotr
        p = r + mu * p
        rdotr = newrdotr
        if rdotr < residual_tol:
            break

    return x, i + 1, rdotr


@torch.no_grad()
def line_search(
    func,
    x_0,
    d_x,
    expected_improvement,
    y_0=None,
    accept_ratio=0.1,
    backtrack_ratio=0.8,
    max_backtracks=15,
    atol=1e-7,
):
    """Perform a linesearch on func with start x_0 and direction d_x."""
    # pylint:disable=too-many-arguments
    if y_0 is None:
        y_0 = func(x_0)

    if expected_improvement >= atol:
        for exp in range(max_backtracks):
            ratio = backtrack_ratio ** exp
            x_new = x_0 - ratio * d_x
            y_new = func(x_new)
            improvement = y_0 - y_new
            # Armijo condition
            if improvement / (expected_improvement * ratio) >= accept_ratio:
                return x_new, expected_improvement * ratio, improvement

    return x_0, expected_improvement, 0



# pylint:disable=missing-module-docstring
from .deterministic import DeterministicPolicy, MLPDeterministicPolicy
from .stochastic import (
    Alpha,
    MLPContinuousPolicy,
    MLPDiscretePolicy,
    MLPStochasticPolicy,
    StochasticPolicy,
)



"""Parameterized stochastic policies."""
from typing import Callable, List

import torch
import torch.nn as nn
from gym.spaces import Box, Discrete

import raylab.torch.nn as nnx
import raylab.torch.nn.distributions as ptd
from raylab.policy.modules.networks.mlp import StateMLP


class StochasticPolicy(nn.Module):
    """Represents a stochastic policy as a conditional distribution module."""

    # pylint:disable=abstract-method

    def __init__(self, params_module, dist_module):
        super().__init__()
        self.params = params_module
        self.dist = dist_module

    def forward(self, obs):  # pylint:disable=arguments-differ
        return self.params(obs)

    @torch.jit.export
    def sample(self, obs, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        params = self(obs)
        return self.dist.sample(params, sample_shape)

    @torch.jit.export
    def rsample(self, obs, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs)
        return self.dist.rsample(params, sample_shape)

    @torch.jit.export
    def log_prob(self, obs, action):
        """
        Returns the log of the probability density/mass function evaluated at `action`.
        """
        params = self(obs)
        return self.dist.log_prob(action, params)

    @torch.jit.export
    def cdf(self, obs, action):
        """Returns the cumulative density/mass function evaluated at `action`."""
        params = self(obs)
        return self.dist.cdf(action, params)

    @torch.jit.export
    def icdf(self, obs, prob):
        """Returns the inverse cumulative density/mass function evaluated at `prob`."""
        params = self(obs)
        return self.dist.icdf(prob, params)

    @torch.jit.export
    def entropy(self, obs):
        """Returns entropy of distribution."""
        params = self(obs)
        return self.dist.entropy(params)

    @torch.jit.export
    def perplexity(self, obs):
        """Returns perplexity of distribution."""
        params = self(obs)
        return self.dist.perplexity(params)

    @torch.jit.export
    def reproduce(self, obs, action):
        """Produce a reparametrized sample with the same value as `action`."""
        params = self(obs)
        return self.dist.reproduce(action, params)

    @torch.jit.export
    def deterministic(self, obs):
        """
        Generates a deterministic sample or batch of samples if the distribution
        parameters are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs)
        return self.dist.deterministic(params)


class MLPStochasticPolicy(StochasticPolicy):
    """Stochastic policy with multilayer perceptron state encoder.

    Args:
        obs_space: Observation space
        spec: Specifications for the encoder
        params_fn: Callable that builds a module for computing distribution
            parameters given the number of state features
        dist: Conditional distribution module

    Attributes:
        encoder: Multilayer perceptron state encoder
        spec: MLP spec instance
    """

    spec_cls = StateMLP.spec_cls

    def __init__(
        self,
        obs_space: Box,
        spec: StateMLP.spec_cls,
        params_fn: Callable[[int], nn.Module],
        dist: ptd.ConditionalDistribution,
    ):
        encoder = StateMLP(obs_space, spec)
        params = params_fn(encoder.out_features)
        params_module = nn.Sequential(encoder, params)
        super().__init__(params_module, dist)

        self.encoder = encoder
        self.spec = spec

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all Linear models in the encoder.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.encoder.initialize_parameters(initializer_spec)


class MLPContinuousPolicy(MLPStochasticPolicy):
    """Multilayer perceptron policy for continuous actions.

    Args:
        obs_space: Observation space
        action_space: Action space
        mlp_spec: Specifications for the multilayer perceptron
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state
    """

    def __init__(
        self,
        obs_space: Box,
        action_space: Box,
        mlp_spec: MLPStochasticPolicy.spec_cls,
        input_dependent_scale: bool,
    ):
        def params_fn(out_features):
            return nnx.PolicyNormalParams(
                out_features,
                action_space.shape[0],
                input_dependent_scale=input_dependent_scale,
            )

        dist = ptd.TransformedDistribution(
            ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1),
            ptd.flows.TanhSquashTransform(
                low=torch.as_tensor(action_space.low),
                high=torch.as_tensor(action_space.high),
                event_dim=1,
            ),
        )
        super().__init__(obs_space, mlp_spec, params_fn, dist)


class MLPDiscretePolicy(MLPStochasticPolicy):
    """Multilayer perceptron policy for discrete actions.

    Args:
        obs_space: Observation space
        action_space: Action space
        mlp_spec: Specifications for the multilayer perceptron
    """

    def __init__(
        self,
        obs_space: Box,
        action_space: Discrete,
        mlp_spec: MLPStochasticPolicy.spec_cls,
    ):
        def params_fn(out_features):
            return nnx.CategoricalParams(out_features, action_space.n)

        dist = ptd.Categorical()
        super().__init__(obs_space, mlp_spec, params_fn, dist)


class Alpha(nn.Module):
    """Wraps a single scalar coefficient parameter.

    Allows learning said coefficient by having it as a parameter

    Args:
        initial_alpha: Value to initialize the coefficient to

    Attributes:
        lob_alpha: Natural logarithm of the current coefficient
    """

    def __init__(self, initial_alpha: float):
        super().__init__()
        self.log_alpha = nn.Parameter(torch.empty([]).fill_(initial_alpha).log())

    def forward(self):  # pylint:disable=arguments-differ
        return self.log_alpha.exp()



"""Rectified Adam implementations

Notes:
    Source::
        https://github.com/LiyuanLucasLiu/RAdam

    Copyright 2019 Liyuan Liu

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0
"""
import math
from typing import Callable, Iterator, Optional, Tuple

import torch
import torch.nn as nn
from torch.optim import Optimizer


class RAdam(Optimizer):
    r"""Rectified Adam

    Reference:
        Liu, Liyuan, et al.
        "On the variance of the adaptive learning rate and beyond."
        arXiv preprint arXiv:1908.03265 (2019).
    """

    def __init__(
        self,
        params: Iterator[nn.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        degenerated_to_sgd: bool = True,
    ):
        # pylint:disable=too-many-arguments
        if lr < 0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        if (
            isinstance(params, (list, tuple))
            and len(params) > 0
            and isinstance(params[0], dict)
        ):
            for param in params:
                if "betas" in param and (
                    param["betas"][0] != betas[0] or param["betas"][1] != betas[1]
                ):
                    param["buffer"] = [[None, None, None] for _ in range(10)]
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            buffer=[[None, None, None] for _ in range(10)],
        )
        super().__init__(params, defaults)

    def step(
        self, closure: Callable[[], torch.Tensor] = None
    ) -> Optional[torch.Tensor]:
        # pylint:disable=too-many-locals,invalid-name,too-many-branches
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                state["step"] += 1
                buffered = group["buffer"][int(state["step"] % 10)]
                if state["step"] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state["step"]
                    beta2_t = beta2 ** state["step"]
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = math.sqrt(
                            (1 - beta2_t)
                            * (N_sma - 4)
                            / (N_sma_max - 4)
                            * (N_sma - 2)
                            / N_sma
                            * N_sma_max
                            / (N_sma_max - 2)
                        ) / (1 - beta1 ** state["step"])
                    elif self.degenerated_to_sgd:
                        step_size = 1.0 / (1 - beta1 ** state["step"])
                    else:
                        step_size = -1
                    buffered[2] = step_size

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(
                            p_data_fp32, alpha=-group["weight_decay"] * group["lr"]
                        )
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(exp_avg, denom, value=-step_size * group["lr"])
                    p.data.copy_(p_data_fp32)
                elif step_size > 0:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(
                            p_data_fp32, alpha=-group["weight_decay"] * group["lr"]
                        )
                    p_data_fp32.add_(exp_avg, alpha=-step_size * group["lr"])
                    p.data.copy_(p_data_fp32)

        return loss


class PlainRAdam(Optimizer):
    """Plain version of RAdam optimizer."""

    def __init__(
        self,
        params: Iterator[nn.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        degenerated_to_sgd: bool = True,
    ):
        # pylint:disable=too-many-arguments
        if lr < 0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        self.degenerated_to_sgd = degenerated_to_sgd
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)

        super().__init__(params, defaults)

    def step(
        self, closure: Callable[[], torch.Tensor] = None
    ) -> Optional[torch.Tensor]:
        # pylint:disable=too-many-locals,invalid-name
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                state["step"] += 1
                beta2_t = beta2 ** state["step"]
                N_sma_max = 2 / (1 - beta2) - 1
                N_sma = N_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(
                            p_data_fp32, alpha=-group["weight_decay"] * group["lr"]
                        )
                    step_size = (
                        group["lr"]
                        * math.sqrt(
                            (1 - beta2_t)
                            * (N_sma - 4)
                            / (N_sma_max - 4)
                            * (N_sma - 2)
                            / N_sma
                            * N_sma_max
                            / (N_sma_max - 2)
                        )
                        / (1 - beta1 ** state["step"])
                    )
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(exp_avg, denom, value=-step_size)
                    p.data.copy_(p_data_fp32)
                elif self.degenerated_to_sgd:
                    if group["weight_decay"] != 0:
                        p_data_fp32.add_(
                            p_data_fp32, alpha=-group["weight_decay"] * group["lr"]
                        )
                    step_size = group["lr"] / (1 - beta1 ** state["step"])
                    p_data_fp32.add_(exp_avg, alpha=-step_size)
                    p.data.copy_(p_data_fp32)

        return loss


class AdamW(Optimizer):
    """Adam optimizer with warmup."""

    def __init__(
        self,
        params: Iterator[nn.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0,
        warmup: int = 0,
    ):
        # pylint:disable=too-many-arguments
        if lr < 0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        defaults = dict(
            lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, warmup=warmup
        )
        super().__init__(params, defaults)

    def step(
        self, closure: Callable[[], torch.Tensor] = None
    ) -> Optional[torch.Tensor]:
        # pylint:disable=too-many-locals,invalid-name
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse gradients,"
                        "please consider SparseAdam instead"
                    )

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

                denom = exp_avg_sq.sqrt().add_(group["eps"])
                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]

                if group["warmup"] > state["step"]:
                    scheduled_lr = 1e-8 + state["step"] * group["lr"] / group["warmup"]
                else:
                    scheduled_lr = group["lr"]

                step_size = (
                    scheduled_lr * math.sqrt(bias_correction2) / bias_correction1
                )

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(
                        p_data_fp32, alpha=-group["weight_decay"] * scheduled_lr
                    )

                p_data_fp32.addcdiv_(exp_avg, denom, value=-step_size)

                p.data.copy_(p_data_fp32)

        return loss



"""Flow implementations.

Each flow is invertible so it can be computed (if training) and inverted
(if evaluating). Each flow also outputs its log of the Jacobian determinant.

References:
    Papamakarios, George, et al. "Normalizing Flows for Probabilistic Modeling
    and Inference." arXiv preprint arXiv:1912.02762 (2019).
"""
from . import masks, utils
from .abstract import (
    CompositeTransform,
    ConditionalTransform,
    InverseTransform,
    Transform,
)
from .affine_constant import ActNorm, AffineConstantFlow
from .coupling import (
    AdditiveCouplingTransform,
    AffineCouplingTransform,
    CouplingTransform,
    PiecewiseRQSCouplingTransform,
)
from .maf import IAF, MAF
from .nonlinearities import (
    AffineTransform,
    SigmoidSquashTransform,
    SigmoidTransform,
    TanhSquashTransform,
    TanhTransform,
)

__all__ = [
    "ActNorm",
    "AdditiveCouplingTransform",
    "AffineConstantFlow",
    "AffineCouplingTransform",
    "AffineTransform",
    "ConditionalTransform",
    "CouplingTransform",
    "IAF",
    "InverseTransform",
    "MAF",
    "masks",
    "PiecewiseRQSCouplingTransform",
    "SigmoidTransform",
    "SigmoidSquashTransform",
    "TanhTransform",
    "TanhSquashTransform",
    "Transform",
]



# pylint:disable=missing-module-docstring
from .action_value import ActionValueCritic
from .q_value import (
    ClippedQValue,
    ForkedQValueEnsemble,
    MLPQValue,
    QValue,
    QValueEnsemble,
)
from .v_value import (
    ClippedVValue,
    ForkedVValueEnsemble,
    HardValue,
    MLPVValue,
    SoftValue,
    VValue,
    VValueEnsemble,
)



"""Parameterized deterministic policies."""
import warnings
from dataclasses import dataclass
from typing import Optional

import torch
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor

import raylab.torch.nn as nnx
from raylab.policy.modules.networks.mlp import StateMLP


class DeterministicPolicy(nn.Module):
    """Continuous action deterministic policy as a sequence of modules.

    If a noise module is passed, it is evaluated on unconstrained actions before
    the squashing module.

    Args:
        encoder: NN module mapping states to 1D features
        action_linear: Linear module mapping features to unconstrained actions
        squashing: Invertible module mapping unconstrained actions to bounded
            action space
        noise: Optional stochastic module adding noise to unconstrained actions
    """

    def __init__(
        self,
        encoder: nn.Module,
        action_linear: nn.Module,
        squashing: nn.Module,
        noise: Optional[nn.Module] = None,
    ):
        super().__init__()
        self.encoder = encoder
        self.action_linear = action_linear
        self.squashing = squashing
        self.noise = noise

    def forward(self, obs: Tensor) -> Tensor:  # pylint:disable=arguments-differ
        """Main forward pass mapping observations to actions."""
        unconstrained_action = self.unconstrained_action(obs)
        return self.squashing(unconstrained_action)

    @torch.jit.export
    def unconstrained_action(self, obs: Tensor) -> Tensor:
        """Forward pass with no squashing at the end"""
        features = self.encoder(obs)
        unconstrained_action = self.action_linear(features)
        if self.noise is not None:
            unconstrained_action = self.noise(unconstrained_action)
        return unconstrained_action

    @torch.jit.export
    def squash_action(self, unconstrained_action: Tensor) -> Tensor:
        """Returns the action generated by the given unconstrained action."""
        return self.squashing(unconstrained_action)

    @torch.jit.export
    def unsquash_action(self, action: Tensor) -> Tensor:
        """Returns the unconstrained action which generated the given action."""
        return self.squashing(action, reverse=True)

    @classmethod
    def add_gaussian_noise(cls, policy, noise_stddev: float):
        """Adds a zero-mean Gaussian noise module to a DeterministicPolicy.

        Args:
            policy: The deterministic policy.
            noise_stddev: Standard deviation of the Gaussian noise

        Returns:
            A deterministic policy sharing all paremeters with the input one and
            with additional noise module before squashing.
        """
        if policy.noise is not None:
            warnings.warn(
                "Adding Gaussian noise to already noisy policy. Are you sure you"
                " called `add_gaussian_noise` on the right policy?"
            )
        noise = nnx.GaussianNoise(noise_stddev)
        return cls(policy.encoder, policy.action_linear, policy.squashing, noise=noise)


@dataclass
class MLPDeterministicPolicySpec(StateMLP.spec_cls):
    """Specifications for deterministic mlp policy network.

    Args:
        units: Number of units in each hidden layer
        activation: Nonlinearity following each linear layer
        layer_norm: Whether to apply layer normalization between each linear layer
            and following activation
        norm_beta: Maximum l1 norm of the unconstrained actions. If None, won't
            normalize actions before squashing function
    """

    norm_beta: float = 1.2

    def __post_init__(self):
        cls_name = type(self).__name__
        assert self.norm_beta > 0, f"{cls_name}.norm_beta must be positive"


class MLPDeterministicPolicy(DeterministicPolicy):
    """DeterministicPolicy with multilayer perceptron encoder.

    The final Linear layer is initialized so that actions are near the origin
    point.

    Args:
        obs_space: Observation space
        action_space: Action space
        mlp_spec: Multilayer perceptron specifications
        norm_beta: Maximum l1 norm of the unconstrained actions. If None, won't
            normalize actions before squashing function.

    Attributes:
        spec_cls: Expected class of `spec` init argument
    """

    spec_cls = MLPDeterministicPolicySpec

    def __init__(
        self,
        obs_space: Box,
        action_space: Box,
        spec: MLPDeterministicPolicySpec,
    ):
        encoder = StateMLP(obs_space, spec)

        action_size = action_space.shape[0]
        if spec.norm_beta:
            action_linear = nnx.NormalizedLinear(
                encoder.out_features, action_size, beta=spec.norm_beta
            )
        else:
            action_linear = nn.Linear(encoder.out_features, action_size)

        action_low, action_high = map(
            torch.as_tensor, (action_space.low, action_space.high)
        )
        squash = nnx.TanhSquash(action_low, action_high)

        super().__init__(encoder, action_linear, squash)

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all Linear models in the encoder.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.encoder.initialize_parameters(initializer_spec)



"""Random LQR problem generators."""
# pylint:disable=invalid-name
from typing import Tuple, Union

import numpy as np
import torch
from numpy.random import Generator
from sklearn.datasets import make_spd_matrix

from .types import LQR, Affine, Box, Quadratic


def box_ddp_random_lqr(
    timestep: float, ctrl_coeff: float, np_random: Generator
) -> Tuple[LQR, Box]:
    # pylint:disable=line-too-long
    """Generate a random, control-limited LQR as described in the Box-DDP paper.

    Taken from `Control-limited differential dynamic programming`_.

    .. _`Control-limited differential dynamic programming`: https://doi.org/10.1109/ICRA.2014.6907001
    """
    # pylint:enable=line-too-long
    assert 0 < timestep < 1

    state_size = np_random.integers(10, 100, endpoint=True)
    ctrl_size = np_random.integers(1, state_size // 2, endpoint=True)

    Fs = _generate_Fs(state_size, ctrl_size, timestep)
    Cs = _generate_Cs(state_size, ctrl_size, timestep, ctrl_coeff)
    bounds = map(torch.from_numpy, (s * np.ones_like(ctrl_size) for s in (-1, 1)))
    return Fs + Cs, bounds


def _generate_Fs(state_size: int, ctrl_size: int, timestep: float) -> Affine:
    F_x = torch.eye(state_size) + timestep * torch.randn(state_size, state_size)
    F_u = torch.randn(state_size, ctrl_size)
    F = torch.cat([F_x, F_u], dim=-1)
    f = torch.zeros(state_size)
    return F, f


def _generate_Cs(
    state_size: int, ctrl_size: int, timestep: float, ctrl_coeff: float
) -> Quadratic:
    dim = state_size + ctrl_size
    C = torch.zeros(dim, dim)

    C_xx = torch.eye(state_size, state_size) * timestep
    C_uu = torch.eye(ctrl_size, ctrl_size) * timestep * ctrl_coeff
    C[:state_size, :state_size] = C_xx
    C[state_size:, state_size:] = C_uu

    c = torch.zeros(dim)
    return C, c


def make_lqr(state_size: int, ctrl_size: int, np_random: Generator) -> LQR:
    """Random LQR generator used in backpropagation-planning.

    Args:
        state_size: Integer size for state
        ctrl_size: Integer size for controls
        np_random: Numpy random number generator

    Source::
        https://github.com/renato-scaroni/backpropagation-planning/blob/master/src/Modules/Envs/lqr.py
    """
    n_dim = state_size + ctrl_size

    F = np_random.normal(size=(state_size, n_dim))
    f = np_random.normal(size=(state_size,))

    C = make_spd_matrix(n_dim)
    c = np_random.normal(size=(n_dim,))

    return tuple(map(torch.Tensor, (F, f, C, c)))


def make_lqr_linear_navigation(
    goal: Union[np.ndarray, Tuple[float, float]], beta: float
) -> Tuple[LQR, Box]:
    """Goal-oriented 2D Navigation task encoded as an LQR.

    Args:
        goal: 2D coordinates of goal position
        beta: Penalty coefficient for control magnitude

    Source::
        https://github.com/renato-scaroni/backpropagation-planning/blob/master/src/Modules/Envs/lqr.py
    """
    goal = np.asarray(goal)
    state_size = ctrl_size = goal.shape[0]

    F = np.concatenate([np.identity(state_size)] * ctrl_size, axis=1).astype("f")
    f = np.zeros((state_size,)).astype("f")

    C = np.diag([2.0] * state_size + [2.0 * beta] * ctrl_size).astype("f")
    c = np.concatenate([-2.0 * goal, np.zeros((ctrl_size,))], axis=0).astype("f")

    bounds = map(torch.from_numpy, (s * np.ones_like(ctrl_size) for s in (-1, 1)))
    lqr = tuple(map(torch.Tensor, (F, f, C, c)))
    return lqr, bounds



"""Collection of invertible transform functions, or Flows."""
import math

import torch
import torch.nn.functional as F
from torch.distributions import constraints
from torch.distributions.transforms import Transform


class TanhTransform(Transform):
    """Transform via the mapping :math:`y = \frac{e^x - e^{-x}} {e^x + e^{-x}}`."""

    domain = constraints.real
    codomain = constraints.interval(-1, +1)
    bijective = True
    sign = +1
    eps = torch.as_tensor(torch.finfo(torch.float32).eps)

    def __eq__(self, other):
        return isinstance(other, TanhTransform)

    def _call(self, x):
        return torch.tanh(x)

    def _inverse(self, y):
        to_log1 = torch.max(1 + y, self.eps)
        to_log2 = torch.max(1 - y, self.eps)
        return (torch.log(to_log1) - torch.log(to_log2)) / 2

    def log_abs_det_jacobian(self, x, y):
        # Taken from spinningup's implementation of SAC
        return 2 * (math.log(2) - x - F.softplus(-2 * x))



# pylint:disable=missing-module-docstring
from abc import ABC, abstractmethod
from typing import List, Union

import torch
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor
from torch.jit import fork, wait

from raylab.policy.modules.actor import Alpha, DeterministicPolicy, StochasticPolicy
from raylab.policy.modules.networks.mlp import StateMLP

from .q_value import ClippedQValue, QValue, QValueEnsemble

MLPSpec = StateMLP.spec_cls


class VValue(ABC, nn.Module):
    """Neural network module emulating a V value function."""

    # pylint:disable=arguments-differ
    @abstractmethod
    def forward(self, obs: Tensor) -> Tensor:
        """Main forward pass mapping obs to V-values.

        Note:
            Output tensor has scalars for each batch dimension, i.e., for a
            batch of 10 observations, the output will have shape (10,).
        """


class MLPVValue(VValue):
    """V-value function with a multilayer perceptron encoder.

    Args:
        obs_space: Observation space
        spec: Multilayer perceptron specifications

    Attributes:
        encoder: NN module mapping observations to logits
        value_linear: Linear module mapping logits to values
    """

    spec_cls = MLPSpec

    def __init__(self, obs_space: Box, spec: MLPSpec):
        super().__init__()

        self.encoder = StateMLP(obs_space, spec)
        self.value_linear = nn.Linear(self.encoder.out_features, 1)

    def forward(self, obs: Tensor) -> Tensor:
        logits = self.encoder(obs)
        return self.value_linear(logits).squeeze(dim=-1)


class VValueEnsemble(nn.ModuleList):
    """A static list of V-value estimators.

    Args:
        v_values: A list of VValue modules
    """

    # pylint:disable=abstract-method
    def __init__(self, v_values: List[VValue]):
        cls_name = type(self).__name__
        assert all(
            isinstance(v, VValue) for v in v_values
        ), f"All modules in {cls_name} must be instances of VValue."
        super().__init__(v_values)

    def forward(self, obs: Tensor) -> List[Tensor]:
        """Evaluate each V estimator in the ensemble.

        Args:
            obs: The observation tensor

        Returns:
            List of `N` output tensors, where `N` is the ensemble size
        """
        # pylint:disable=arguments-differ
        return self._state_values(obs)

    def _state_values(self, obs: Tensor) -> List[Tensor]:
        return [m(obs) for m in self]

    @staticmethod
    def clipped(outputs: List[Tensor]) -> Tensor:
        """Returns the minimum V-value of an ensemble's outputs."""
        mininum, _ = torch.stack(outputs, dim=0).min(dim=0)
        return mininum


class ForkedVValueEnsemble(VValueEnsemble):
    """Ensemble of V-value estimators with parallelized forward pass."""

    # pylint:disable=abstract-method

    def _state_values(self, obs: Tensor) -> List[Tensor]:
        futures = [fork(m, obs) for m in self]
        return [wait(f) for f in futures]


class SoftValue(VValue):
    """V-value computed from stochastic policy, Q-value, and entropy bonus."""

    def __init__(
        self,
        policy: StochasticPolicy,
        q_value: Union[QValue, QValueEnsemble],
        alpha: Alpha,
        deterministic: bool = False,
    ):
        super().__init__()
        if isinstance(q_value, QValueEnsemble):
            # Treat everything as if single value
            q_value = ClippedQValue(q_value)
        self.q_value = q_value

        self.policy = policy
        self.alpha = alpha

        self.deterministic = deterministic

    def forward(self, obs: Tensor) -> Tensor:
        if self.deterministic:
            action, logp = self.policy.deterministic(obs)
        else:
            action, logp = self.policy.rsample(obs)

        alpha = self.alpha()
        entropy_bonus = -alpha * logp
        action_value = self.q_value(obs, action)
        return action_value + entropy_bonus


class HardValue(VValue):
    """V-value computed from deterministic policy and Q-value."""

    def __init__(
        self, policy: DeterministicPolicy, q_value: Union[QValue, QValueEnsemble]
    ):
        super().__init__()
        self.policy = policy

        if isinstance(q_value, QValueEnsemble):
            # Treat everything as if single value
            q_value = ClippedQValue(q_value)
        self.q_value = q_value

    def forward(self, obs):
        return self.q_value(obs, self.policy(obs))


class ClippedVValue(VValue):
    """Minimum of an ensemble of state-value functions."""

    def __init__(self, v_values: VValueEnsemble):
        super().__init__()
        self.v_values = v_values

    def forward(self, obs: Tensor) -> Tensor:
        values = self.v_values(obs)
        minimum, _ = torch.stack(values, dim=0).min(dim=0)
        return minimum
